#ifndef BCPGRAPH_HPP_
#define BCPGRAPH_HPP_
#define bcpengine_hpp_version "$Revision: 1.30 $"



#include "GlobalSettings.hpp"

#include "AtomicTypes.hpp"
#include "SharedClause.hpp"
//#include "VecOfVecs.hpp"
#include "VariableInfos.hpp"
#include "Logging.hpp"
#include "Heap.hpp"
#include "BitVector.hpp"
#include "SiblingInfo.hpp"
#include "Satistics.hpp"
#include "Watching.hpp"
#include "SimplyFire.hpp"
#include "ElimClses.hpp"


using std::pair; using std::make_pair;
using std::vector;


enum BeyondUP{NoBUP=0,PessimisticBUP=1,OptimisticBUP=2};
enum HyperBinRes{NoHyperBin=0,BiereHyperBin=1};//,AdvancedHyperBin=2};
enum OnTheFlySubsum{NoOnTheFlySub,DoOnTheFlySub};


#define BCPETEMP template<class Decision,class SInfo, class ClsRegistry>
#define BCPECLASS BcpEngine<Decision,SInfo,ClsRegistry>


struct UkwnVisitor{
	const BinVarInfo& v_info;
	UkwnVisitor(const BinVarInfo& v_i):v_info(v_i){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		return TrBool(v_info.litUkwn(tgt)) ;}
};

struct FalseVisitor:public UkwnVisitor{
	FalseVisitor(const BinVarInfo& v_i):UkwnVisitor(v_i){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		LOG3("Visit " << tgt.toDIMACS());
		return TrBool(v_info.litFalse(tgt));}
};

struct TwiceVisitor{
	BinVarInfo& v_info; uint32_t cnt; const Literal rch;
	TwiceVisitor(BinVarInfo& v_i,const Literal& r):v_info(v_i),cnt(0),rch(r){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		v_info.clearFlag(rch);
		if(tgt==rch && ++cnt > 1){return UNKNOWN_TrBool;}
		return TRUE_TrBool;}
};

struct CollectUkwnVisitor: public UkwnVisitor{
	vector<pair<Literal,Literal> >	common;
	CollectUkwnVisitor(const BinVarInfo& v_i):UkwnVisitor(v_i){}
	TrBool operator()(const Literal& tgt, const Literal& src){
		if(v_info.litKnwn(tgt)) return FALSE_TrBool;
		common.push_back(std::make_pair(tgt,src));
		return TRUE_TrBool;
	}
};

struct CollectFalseVisitor: public CollectUkwnVisitor{
	CollectFalseVisitor(const BinVarInfo& v_i):CollectUkwnVisitor(v_i){}
	TrBool operator()(const Literal& tgt, const Literal& src){ asrt(v_info.litFalse(src));
		LOG3("Visit " << tgt.toDIMACS());
		if(!v_info.litFalse(tgt)||v_info.litFix(tgt)) return FALSE_TrBool;
		asrt(src==FALSE_LIT || v_info.litAssLev(tgt)==v_info.litAssLev(src));
		common.push_back(std::make_pair(tgt,src));
		return TRUE_TrBool;
	}
};

template<class ClsReg,class SInfo>
struct SolvState{
	SInfo*			s_info;
	Satistics*		stats;
	VarVal			vals;
	BinVarInfo 		vinf;
	Watching		wtch;
	ClsInfo			cinf;
	ClsReg			cls_reg;
	SolvState(Satistics* sta,SInfo* sif):s_info(sif),stats(sta),
			vinf(vals),cinf(sta),cls_reg(vinf,cinf,wtch){}

	inline void createClone(const SolvState<SimplyFire,LeaderSiblingInfo>& s);
	inline void addBinAndSend(const Literal& l1, const Literal& l2);
	inline ClsPointer	linkCls(SharedClause* cls,const uint32_t& sz, const Literal& w0, const Literal& w1,
				const UserMask& recomm_for,const uint32_t& loc_lbd_msk = 0);
	inline SharedClause* ensureClsFix(const ClsPointer& clp, const uint32_t& sz);
	bool varReplFixOrElim(const Variable& v){return vinf.varFix(v) || vinf.varRepl(v) || vinf.varElim(v);}
	bool litReplTrueOrElim(const Literal& l){return vinf.litTrue(l) || vinf.litRepl(l) || vinf.litElim(l);}
};


BCPETEMP
class BcpEngine{
public:
	typedef Decision  					decision_t;
	typedef typename decision_t::act_t	act_t;
	typedef SInfo						sinfo_t;

	SolvState<ClsRegistry,SInfo>		sstate;

	/** Assignment stack; stores all assignments made in the order they were made*/
	vector<Literal>			assign_stack;

	//will be set by subclass
	decision_t*				decision;


protected:
	const BeyondUP 			bey_up;
	const HyperBinRes 		hb_res;
	const OnTheFlySubsum 	otf_sub;
	char*					invalid;

	/** Head of queue - as index into the assign_stack */

	uint32_t				queue_begin;

	/** Indices for different decision levels in 'assign_stack'
	 * decision_level_end[l] states that all assignments in assign_stack[0,l) are made above or at decision level l*/
	vector<uint32_t>		decision_level_end;


	BitVector<Literal>		lit_flags;
	vector<Literal>			some_lits;
	//space to save some literals (for instance in firstDecisionCheckBothPol(), reduceSelfImpl())
	vector<Literal>	 		tmp_literal_store;

	//conflict related information
	vector<Literal>			lemma; //to generate the lemma
	ClsPointer				conflicting; //for onthefly-subsumption
	BitVector<uint32_t>		level_flags; //involved levels
	uint32_t				confl_levs; //number of involved levels
	const bool				bump_glue; //extra bump literals
	vector<Literal>			confl_lits;

	//beyond unit propagation
	vector<ClsPointer>		beyUP_cands; //clauses that might lead to beyond-bcp decisions
	uint32_t				beyUP_head; //index into beybcp_cands

	uint32_t				units_sent; //index into assign_stack

public:
	//creation of instances
	inline 				BcpEngine(Satistics* sta,decision_t* d,sinfo_t* sinfo,
						const bool glue,const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub);
	inline void 		createClone(const SolvState<SimplyFire,LeaderSiblingInfo>& from, const vector<Literal>& units); //before parallel start for troopers
	inline 				~BcpEngine();
	inline bool			reserveForVariables(uint32_t s);
	inline void			reserveForClauses(uint32_t s);

	//information
	uint32_t 			nAssignments()		const	{return assign_stack.size();}
	uint32_t 			nVars()				const	{return sstate.vinf.varsCount();}
	uint32_t 			nClauses()			const	{return sstate.cinf.clsCount();}


	//show the result
	inline void			printModel();
	inline Literal		getModelVal(const Variable& v);
	inline bool			oneKnwnTrue(ElimCls* eli);
	inline void			printDIMACS(const string& file,const TrBool& result);
	inline void			getModel(vector<bool>& var_vals);
	inline void			uneliminate();
	bool				setInvalid(const char* why)	{if(invalid==NULL){invalid = (char*)why;} return false;}
	bool				ok()				const	{return invalid == NULL;}
	char* 				whyInvalid()		const	{return invalid;}


	//where are we
	inline void			newDecisionLevel();
	bool				currLemmaLong()		const	{return lemma.size() > 2;}
	uint32_t 			decisionLevel()		const	{return decision_level_end.size(); }
	uint32_t 			decisionsAtLevel0()	const	{return decision_level_end.empty()?assign_stack.size():decision_level_end[0]; }
	bool				openBCPImpls()		const	{return queue_begin != assign_stack.size();}
	bool				openCompImpls()		const	{return !beyUP_cands.empty();}

	//set flags for conflicts, decisions and some things else
	bool				isSet(const Literal& l) 	const 	{return lit_flags.isSet(l);}
	void				setFlag(const Literal& l)			{lit_flags.setFlag(l);}
	void				clearFlag(const Literal& l)			{lit_flags.clear(l);}
	void				zeroFlags(const Literal& l)			{lit_flags.carelessClear(l);}
	void				zeroFlaggs(const vector<Literal>& vec){for(uint32_t i=vec.size();i--;){zeroFlags(vec[i]);}}
	void				zeroFlaggs(Literal lits[],uint32_t sz){while(sz--){zeroFlags(lits[sz]);}}
	bool				levelSet(const uint32_t& l) const 	{return level_flags.isSet(l);}
	void				setLevel(const uint32_t& l)			{confl_levs += level_flags.setFlagNew(l);}
	void				delLevel(const uint32_t& l)			{level_flags.carelessClear(l);}


	static const int32_t BCP_SUCCESS 		= -1;

private:
	inline int32_t		bcpInitConflict		(ClsPointer clp);
	inline int32_t		bcpInitConflictUnit	(const Literal& lit);
	inline int32_t		bcpFindNewWatch(ClsWrp& cw, const WtchEle& we, const Literal& false_lit,WIter<WtchEle>& wi);
	inline uint32_t		filterFalse(vector<Literal>& false_out, SharedClause* cls, uint32_t& levs,Literal& latest);
	inline uint32_t 	filterUkwnIfRooted(SharedClause* cls, vector<Literal>& lits_out);

	inline Literal 		bcpBeyondUP(SharedClause* cls);

	template<class VisiColl,class VisiMark>
	inline void			collectCommon(Literal lits[],uint32_t lsz, vector<Literal>& common_out);
protected:
	template<class Visitor>
	inline void 		BFS(const Literal& start, Visitor& visi, const bool visit_start, const bool clear_flags);
	inline void			clearBFSFlags();

	//different propagation
	inline Literal		enqueueBinBCP				(const Literal& lit,const Literal& root);
	inline int32_t		bcpLong();
	inline int32_t 		bcpBeyondUP();
	inline bool			enqueue(const Literal& lit, uint32_t lev);
	inline bool			enqueue(const Literal& lit, uint32_t lev,const ClsPointer& clp,const uint32_t lbd);
	inline bool			enqueue(const Literal& lit, uint32_t lev,const Literal& r);

//learning
	inline bool			handleConflict(int32_t bcp_res);
	inline void			backjumpKeepLevel   (uint32_t level);
	inline void 		backjumpZero();
	inline bool			firstUIP			(int32_t nbr_pivot_lits);
	inline int32_t		markConflictLit(const Literal& false_lit,const int32_t& dec_level);
	inline uint32_t 	strengthenClsSetConflicting(SharedClause* cls,Literal p_lit);
	inline uint32_t		backtrackLevel(Literal& w);
	inline void			lemmaReduction		(const bool only_clear);
	inline void			unsetConfLemmaFlags	();
	inline void			bumpGlueIf();
	inline bool			needLitGraphCheck	  	(const Literal& lit);
	inline bool			needLitGraphCheckHelper(const Literal& lit);
	inline bool			needLitShortsCheck(const Literal& lit);


//import and send
	inline void			sendUnits();

	inline Literal		importBinImpliesUnit(Literal a, const Literal& b);
	inline void			findWatchers(SharedClause* cls, Literal& w1, Literal& w2);
	inline bool			imprtLong(SharedClause* cls,uint32_t max_lbd, uint32_t balance_percent, act_t act);
	inline bool			import(uint32_t max_lbd,uint32_t balance_percent);
	inline bool 		import(){return import(sstate.s_info->getMaxImportLBD(), sstate.s_info->getMaxTrueLitsPermil());}

//apply lit replacement and var elimination
	inline bool			applyLitReplsAndRemoves(const vector<Message>& repls_removes);
	inline bool 		replaceVar(const Literal& die, const Literal& survive,vector<Literal>& units_out,
							vector<Literal>& update_bins,	vector<ClsPointer>& update_cls);
	inline void			applyVarElim(const Variable& var, vector<Literal>& lits_to_update);
	inline void			updateReplsInBins(vector<Literal>& lits, vector<Literal>& units_out);
	inline SharedClause* createReplacedIf(SharedClause* cls,uint32_t& sz_out, uint32_t& good_percent_out,act_t border);

//bfs
	inline void			collectCommonFalse(Literal lits[],uint32_t lsz, vector<Literal>& common_out){
						return collectCommon<CollectFalseVisitor,FalseVisitor>(lits,lsz,common_out);}
	inline void			collectCommonUkwn(Literal lits[],uint32_t lsz, vector<Literal>& common_out){
						return collectCommon<CollectUkwnVisitor,UkwnVisitor>(lits,lsz,common_out);}

	inline Literal 		checkSatByLit(const SharedClause* cls)const;


protected: //cleaning
	void clsGC(const uint32_t new_sz);
	//bool arcRedundant(Literal from, Literal to);
	inline void clearWatches(const uint32_t max_lbd);
	inline void	lockReasons();
private:
	inline void clearWatches( const Literal& lit, const uint32_t max_lbd);
	inline void clearBins(const Literal& lit);
	inline void wasteAllWatched( const Literal& lit);


public:
//Debug
	inline bool 		asrtCheckSatisfied(SharedClause* cl) const;
	inline bool			asrtAllLitsFlagged(const vector<Literal>& lits);
	inline bool			asrtClsUnit(SharedClause* cls,const Literal& l)const;
	//inline bool			asrtClsSorted(SharedClause* cls) const;
	//inline bool			asrtSortedAndFlagged(const Literal lits[],const uint32_t lsz)const;
	inline bool			asrtBinPropagatedIf(const Literal& lit);
	inline void			asrtPrintCurrResolvent() const ;
	inline bool			asrtResolventSize(uint32_t sz) const ;
	inline bool			asrtAllUnitsSent() const;
	inline bool			asrtNoReplNoFixLits(SharedClause* cls);
	inline bool			asrtContaisRepl(SharedClause* cls);
	inline bool 		asrtNoClsContainsReplOrElim();
	inline bool			asrtAllLinkedClsesSAT(const vector<bool>& vals);


};





BCPETEMP
BCPECLASS::BcpEngine(Satistics* sta, decision_t* d,sinfo_t* sinfo,
		const bool glue, const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub)
:
	sstate						(sta,sinfo),
	decision					(d),
	bey_up						(beyup),
	hb_res						(hbres),
	otf_sub						(otfsub),
	invalid						(NULL),
	queue_begin					(1),
	confl_levs					(0),
	bump_glue					(glue),
	beyUP_head					(0),
	units_sent					(0)
{
	asrt(decision != NULL); asrt(sstate.stats!= NULL); asrt(sstate.s_info!=NULL);
	reserveForVariables(1);
	sstate.vinf		.setValueDecision(TRUE_LIT,0);
	assign_stack	.push_back(TRUE_LIT);
	decision		->setVarInfo(&sstate.vinf);
	level_flags		.ensureSize(64,false);
}




BCPETEMP
void	BCPECLASS::createClone(const SolvState<SimplyFire,LeaderSiblingInfo>& from,const vector<Literal>& units){
	asrt(units.size()>0); asrt(units[0]==TRUE_LIT);
	asrt(assign_stack.size()==1); asrt(assign_stack[0]==TRUE_LIT);

	sstate					.createClone(from);
	assign_stack			= units;
	units_sent 				= assign_stack.size();
	decision				->ensureSize(sstate.stats->nbr_vars);
	lit_flags				.ensureSize(sstate.stats->nbr_vars*2,true);

	asrt(assign_stack.size()==units.size());
	LOG3("Unit clauses initialised");
}


BCPETEMP
BCPECLASS::~BcpEngine(){
	ClsPointer iter = sstate.cinf.lastValid();
	while(!iter.isNull()){
		ClsWrp& cw = sstate.cinf.get(iter);asrt(cw.getCls() != NULL);
		//SharedClause::release(cw.getCls(),sstate.s_info->user_bit,"Solver destruct");
		SharedClause::release(cw.getCls(),sstate.s_info->user_bit);
		iter = sstate.cinf.nextValidBefore(iter);
	}
	LOG3("~BcpEngine");
}


template<class ClsReg,class SInfo>
void SolvState<ClsReg,SInfo>::createClone(const SolvState<SimplyFire,LeaderSiblingInfo>& s){
	vinf					.createClone(s.vinf);
	wtch					.createClone(s.wtch);
	cinf					.createClone(s.cinf);
	stats->nbr_vars 		= s.stats->nbr_vars;
	stats->nbr_bin_clses 	= s.stats->nbr_bin_clses;
	stats->nbr_lf_cls		= s.stats->nbr_lf_cls;
	stats->nbr_lf_lits		= s.stats->nbr_lf_lits;
}

template<class ClsReg,class SInfo>
void SolvState<ClsReg,SInfo>::addBinAndSend(const Literal& l1, const Literal& l2){
	asrt(l1 !=l2); asrt(!vinf.varElim(l1.toVar())); asrt(!vinf.varElim(l2.toVar()));
	s_info->sendBinCls(l1,l2);
	cls_reg.registerBin(l1,l2);
	wtch.addBinCls(l1,l2);
}

template<class ClsReg,class SInfo>
SharedClause* SolvState<ClsReg,SInfo>::ensureClsFix(const ClsPointer& clp, const uint32_t& sz){
	ClsWrp& cw 			= cinf.get(clp);
	SharedClause* cls	= cw.getCls();
	if(cls->isFix())	return cls;

	SharedClause* ncls  = SharedClause::copyFrom(s_info->all_users_bits,false,cls,sz);
	LOG2("Silent replace at " << clp.idx() << " size " << sz << " old " << (uint64_t)cls << " new " << (uint64_t)ncls);
	cw.silentReplace(ncls,0);

	s_info->sendLongCls(ncls,ncls->getUsers());

	//LOG0("Send cls " << ((uint64_t)cls));
	//cls->setWaste("Silent replace of clause to make it fix size ", sz);
	//SharedClause::release(cls,s_info->user_bit,"Ensure fix replaced this");

	cls->setWaste();
	SharedClause::release(cls,s_info->user_bit);
	return ncls;
}


template<class ClsReg,class SInfo>
ClsPointer	SolvState<ClsReg,SInfo>::linkCls(SharedClause* cls,const uint32_t& sz, const Literal& w0, const Literal& w1,
		const UserMask& recomm_for,const uint32_t& loc_lbd_msk){
	asrt(cls != NULL);asrt(w0 != w1); asrt(!vinf.litFix(w0)); asrt(!vinf.litFix(w1));
	LOG3("Link shared clause with watchers " << w0.toDIMACS()<<","<<w1.toDIMACS());

	ClsPointer clp = cinf.addCls(w0,w1,cls,sz,cls->getOrigLBD() | loc_lbd_msk); asrt(cinf.get(clp).getOtherW(w0) == w1);
	asrt(!clp.isNull());

	wtch.addCls(w0,clp,w1);
	wtch.addCls(w1,clp,w0);


	if(recomm_for.hasAnyUser()) { //LOG0("Send cls " << ((uint64_t)cls));
		s_info->sendLongCls(cls,recomm_for);
	}//else{LOG0("Only link cls " << ((uint64_t)cls));}

	cls_reg.registerClause(clp,cls);

	return clp;
}


BCPETEMP
bool		BCPECLASS::enqueue(const Literal& lit, uint32_t lev){
	LOG3("Enqueue dec " << lit.toDIMACS() << " at level " << lev);
	asrt(sstate.vinf.litUkwn(lit)); asrt(lev<level_flags.size()); asrt(!sstate.vinf.litRepl(lit));
	sstate.vinf			.setValueDecision(lit,lev);
	asrt(sstate.vinf.litAssLev(lit) == lev);	asrt(sstate.vinf.varKnwn(lit.toVar()));
	asrt(sstate.vinf.varHasValue(lit.toVar(),lit.isPos()));	asrt(sstate.vinf.varIsDecision(lit.toVar()));
	return enqueueBinBCP(lit,lit.flipped())==INVALID_LIT;}

BCPETEMP
bool		BCPECLASS::enqueue(const Literal& lit, uint32_t lev,const ClsPointer& clp,const uint32_t lbd){
	LOG3("Enqueue long " << lit.toDIMACS() << " at level " << lev << " lbd " << lbd);
	asrt(sstate.vinf.litUkwn(lit));asrt(lev<level_flags.size()); asrt(asrtClsUnit(sstate.cinf.get(clp).getCls(),lit)); asrt(lbd<=SharedClause::LBD_MAX);
	asrt(SharedClause::LBD_MAX == BDReason::MAX_LBD);

	sstate.vinf		.setValueReason(lit,lev,clp,lbd);
	//cls			->incPropCnt();

	//sstate.cinf			.setUsed(clp);
	asrt(sstate.vinf.litAssLev(lit) == lev);		asrt(sstate.vinf.varKnwn(lit.toVar())); asrt(!sstate.vinf.litRepl(lit));
	asrt(sstate.vinf.varHasValue(lit.toVar(),lit.isPos()));	asrt(sstate.vinf.varHasReason(lit.toVar(),clp));

	return enqueueBinBCP(lit,lit.flipped())==INVALID_LIT;}

BCPETEMP
bool		BCPECLASS::enqueue(const Literal& lit, uint32_t lev,const Literal& r1){
	LOG3("Enqueue bin " << lit.toDIMACS() << " at level " << lev << " reason " << r1.toDIMACS()	);
	asrt(sstate.vinf.litUkwn(lit)); asrt(sstate.vinf.litFalse(r1)); asrt(lev == decisionLevel()); asrt(lev<level_flags.size());

	const Literal root = sstate.vinf.litRoot(r1);

	asrt(sstate.vinf.litFalse(root)); asrt(!sstate.vinf.litRepl(root));
	sanity(sstate.wtch.asrtPathExists(root.flipped(),r1.flipped(),sstate.vinf)); sanity(sstate.wtch.asrtPathExists(r1,root,sstate.vinf));
	asrt(sstate.vinf.litAssLev(r1)==lev); asrt(sstate.vinf.litAssLev(root)==lev); asrt(!sstate.vinf.litRepl(lit));

	sstate.vinf			.setValueReason(lit,lev,r1, root);

	asrt(sstate.vinf.litAssLev(lit) == lev);		asrt(sstate.vinf.varKnwn(lit.toVar()));
	asrt(sstate.vinf.varHasValue(lit.toVar(),lit.isPos()));	asrt(sstate.vinf.varGetBinReason(lit.toVar()) == r1);
	asrt(sstate.vinf.litFalse(sstate.vinf.litRoot(lit.flipped())));

	return enqueueBinBCP(lit,root)==INVALID_LIT;}




BCPETEMP
int32_t	BCPECLASS::bcpLong(){
	TIME_ON(stats->bcp_time);
	WtchEle we;	 int32_t find_res = BCP_SUCCESS; //Literal otherw = INVALID_LIT;

	while(queue_begin < assign_stack.size()){
		LOG3("BCP with q " << queue_begin << " and stack " << assign_stack.size());

		++sstate.stats->nbr_bcp;
		const Literal& false_lit 	= assign_stack[queue_begin++].flipped(); asrt(!sstate.vinf.litRepl(false_lit)); asrt(!sstate.vinf.varElim(false_lit.toVar()));
		WIter<WtchEle> wi			= sstate.wtch.getWtchOfW(false_lit);


		while(wi.getNext(we)){

			if(sstate.vinf.litTrue(we.other)){	wi.keep(we);continue;}

			ClsWrp& cw	= sstate.cinf.get(we.clp);

			we.other	= cw.getOtherW(false_lit); asrt(we.other != false_lit);
			if(sstate.vinf.litTrue(we.other)){	wi.keep(we); continue;}

			find_res = bcpFindNewWatch(cw,we,false_lit,wi);
			if(find_res == BCP_SUCCESS) continue;

			// No new watched found
			asrt(sstate.vinf.litFalse(we.other) || find_res==1);

			//conflict
			asrt(find_res != BCP_SUCCESS);
			while(wi.getNext(we)){wi.keep(we);}
			sstate.wtch.finish(false_lit,wi);
			TIME_OFF(sstate.stats->bcp_time);
			return find_res;
		}
		sstate.wtch.finish(false_lit,wi);
	}

	TIME_OFF(sstate.stats->bcp_time);
	return BCP_SUCCESS;
}

BCPETEMP
int32_t	BCPECLASS::bcpFindNewWatch(ClsWrp& cw, const WtchEle& we,const Literal& false_lit,WIter<WtchEle>& wi){

	LOG3("Search new watcher for cls " << we.clp.idx() << " for lit " << false_lit.toDIMACS() << " otherw is " << we.other.toDIMACS());
	const uint32_t dec_l 	= decisionLevel(); asrt(cw.getCls() != NULL);

	SharedClause* cls	= cw.getCls();
	ClsLit* beg			= cls->beg();
	ClsLit* iter		= cw.addOffset(beg);
	ClsLit* strt		= iter;


	do{
		Literal cand(iter->lit);

		if(!sstate.vinf.litFalse(cand) && cand != we.other){
			cw		.setWatchers(we.other,cand);
			sstate.wtch	.addCls(cand,we.clp,we.other);
			cw.keepOffset(beg,iter);
			LOG3("Found new watcher " << cand.toDIMACS() << " for " << false_lit.toDIMACS());

			if(bey_up >NoBUP
				&& sstate.vinf.litUkwn(cand)
				&& (sstate.vinf.litRoot(cand)==sstate.vinf.litRoot(we.other))
				&& (bey_up==OptimisticBUP ||sstate.vinf.litRootUkwn(cand))){
					cw.setBUP(true);
					beyUP_cands.push_back(we.clp);
			}

			return BCP_SUCCESS;
		}

		iter = (iter->has_next) ? (iter+1) : beg;

	}while(iter != strt);

	wi.keep(we);

	LOG3("No new watcher found");


	if(sstate.vinf.litFalse(we.other)){
		int32_t nbr_pivot_lits 	= bcpInitConflict(we.clp);
		FOR_LITS_STRT(cls,curr);
			nbr_pivot_lits += markConflictLit(curr,dec_l);
		FOR_LITS_END;

		cls->incConflCnt();
		asrt(nbr_pivot_lits != 1);
		LOG2("Marked " << nbr_pivot_lits << " for FUIP ");
		return nbr_pivot_lits;
	}

	asrt(sstate.vinf.litUkwn(we.other));
	LOG3("BCP clause became unit " << we.other.toDIMACS());

	if(dec_l !=0){
		const uint32_t max_lev 	= cw.getLBD();
		Literal root 			= hb_res>NoHyperBin? sstate.vinf.litRoot(false_lit) :FALSE_LIT;
								asrt(sstate.vinf.litFalse(root)); asrt(hb_res==NoHyperBin||sstate.vinf.litAssLev(root)==dec_l);

		sanity(sstate.wtch.asrtPathExists(false_lit,root,sstate.vinf));
		sanity(sstate.wtch.asrtPathExists(root.flipped(),false_lit.flipped(),sstate.vinf));

		uint32_t levs = 0; uint32_t lev; level_flags.setFlag(dec_l);
		FOR_LITS_STRT(cls,lit);

			if(lit==we.other) 				continue;
			lev								= sstate.vinf.litAssLev(lit);
			if(lev==0)						continue;	asrt(sstate.vinf.litFalse(lit)); asrt(level_flags.size()>decision_level_end.size()); asrt(lev<level_flags.size());

			levs 							+= level_flags.setFlagNew(lev);
			if(sstate.vinf.litRoot(lit)!=root){ 	root = FALSE_LIT;}
											asrt(root==FALSE_LIT || sstate.vinf.litAssLev(root)==dec_l);
											sanity(root==FALSE_LIT||sstate.wtch.asrtPathExists(root.flipped(),lit.flipped(),sstate.vinf));

			if(levs>=max_lev&&root==FALSE_LIT)break;

		FOR_LITS_END;

		if(levs<max_lev) {asrt(!level_flags.isSet(0)); LOG2("Change lbd from "<<max_lev << " to " << levs);
			cw.setLBD(levs);}

		FOR_LITS_STRT(cls,l);
			if(l!=we.other){delLevel(sstate.vinf.litAssLev(l));}
		FOR_LITS_END;
		delLevel(dec_l);
		sanity(level_flags.allCleared());

		if(root!=FALSE_LIT){
			LOG2("Found root " <<root.toDIMACS() << " could create bin with " << we.other.toDIMACS());
			asrt(sstate.vinf.litFalse(root)); asrt(we.other != root); asrt(we.other.flipped() != root); asrt(some_lits.empty());
			asrt(sstate.vinf.litAssLev(root)==dec_l); asrt(!sstate.vinf.litFix(root)); asrt(!sstate.vinf.litRepl(root));asrt(hb_res!=NoHyperBin);

			sstate.stats->nbr_hyper_bin_res++;

			LOG2("Add bin " << we.other.toDIMACS() << " with root " << root.toDIMACS());
			sstate.addBinAndSend(we.other, root);

			if(root==false_lit){
				sstate.stats	->nbr_hyper_bin_self_sub++;
				cls		->setWaste();
				//cls		->setWaste("Hyper bin self sub root equals false " , root.toDIMACS());
			}

			return enqueue(we.other,dec_l,root) ? BCP_SUCCESS: bcpInitConflictUnit(we.other.flipped());

			/*Literal lx = root; uint32_t x = 0;
			if(hb_res==AdvancedHyperBin){
				collectCommonFalse(flits,fidx,some_lits); asrt(!some_lits.empty());
				lx = some_lits[x++];

				sstate.stats->nbr_hyper_bin_impr+= (some_lits.size()-1);
				LOG2("May add " << some_lits.size()<<" bins by root check");
			}

			while(true){
				LOG2("Add bin " << otherw.toDIMACS() << ", " << lx.toDIMACS() << " with root " << root.toDIMACS());
				asrt(sstate.vinf.litFalse(lx));asrt(dec_l==sstate.vinf.litAssLev(lx)); asrt(sstate.vinf.litAssLev(sstate.vinf.litRoot(lx))==dec_l);
				sanity(asrtBinPropagatedIf(lx.flipped()));

				addBinAndSend(we.other, lx);

				if(lx==false_lit){
					sstate.stats	->nbr_hyper_bin_self_sub++;
					cls		->setWaste();}

				if(x>=some_lits.size())	break;
				lx = some_lits[x++];
			}

			some_lits.clear();*/
		}

	}
	return enqueue(we.other,dec_l, we.clp,cw.getLBD()) ?
				BCP_SUCCESS : bcpInitConflictUnit(we.other.flipped());
}


BCPETEMP
uint32_t	BCPECLASS::filterFalse(vector<Literal>& false_out, SharedClause* cls, uint32_t& levs,Literal& latest){
	asrt(levs==0); sanity(level_flags.allCleared()); asrt(lit_flags.allCleared());
	uint32_t hlev =0; Literal root = TRUE_LIT; const uint32_t dlev = decisionLevel();
	level_flags.setFlag(dlev);

	FOR_LITS_STRT(cls,lit);

		if(!sstate.vinf.litFalse(lit))	continue; asrt(!sstate.vinf.litTrue(lit));
		const uint32_t lev	= 		sstate.vinf.litAssLev(lit);
		if(lev==0)					continue;
		if(root==TRUE_LIT){			root = sstate.vinf.litRoot(lit); asrt(false_out.empty());}
		else if(sstate.vinf.litRoot(lit) != root){root=FALSE_LIT;}

		levs 						+= level_flags.setFlagNew(lev); asrt(sstate.vinf.litAssLev(lit)==lev);
		if(lev>=hlev)				{latest=lit;hlev=lev;}
		false_out					.push_back(lit);

	FOR_LITS_END;

	asrt(!level_flags.isSet(0)); asrt(!false_out.empty()); asrt(root != TRUE_LIT);

	uint32_t fls = false_out.size(); delLevel(dlev);
	for(uint32_t i=fls;i--;){delLevel(sstate.vinf.litAssLev(false_out[i]));}
	sanity(level_flags.allCleared()); asrt(lit_flags.allCleared());


	if(fls==1){asrt(latest != FALSE_LIT); 	return max32;}
	if(root!=FALSE_LIT){latest=root;		return max32;}
	return fls;
}



BCPETEMP
uint32_t BCPECLASS::filterUkwnIfRooted(SharedClause* cls, vector<Literal>& lits_out){
	Literal root; asrt(root==INVALID_LIT);

	FOR_LITS_STRT(cls,lit);
		if(sstate.vinf.litTrue(lit))	{	return 0;}
		if(sstate.vinf.litFalse(lit)){	continue;}

		if(root==INVALID_LIT){	root = sstate.vinf.litRoot(lit);}
		else if(root != sstate.vinf.litRoot(lit)){ return 0;}

		lits_out.push_back(lit);
	FOR_LITS_END;

	return lits_out.size();
}


BCPETEMP
Literal	BCPECLASS::bcpBeyondUP(SharedClause* cls){
	vector<Literal> ukwn;
	uint32_t ukwn_lits	= filterUkwnIfRooted(cls,ukwn);

	if(ukwn_lits==0)	{sstate.stats->nbr_comp_bcp_cheap_fail++; 	return INVALID_LIT;}

	asrt(some_lits.empty());
	collectCommonUkwn(ukwn.begin().base(),ukwn_lits,some_lits);
	if(some_lits.empty())			{sstate.stats->nbr_comp_bcp_exp_fail++; 	return INVALID_LIT;}

	const uint32_t dec_l 			= decisionLevel();
	if(dec_l==0){
		LOG2("Resolution during Comp BCP creating "<< some_lits.size() << " unit clauses" );
		for(uint32_t i=some_lits.size();i--;){
			Literal imp 		= some_lits[i];
			if(sstate.vinf.litKnwn(imp)) continue;
			LOG3("Enqueue lit " << imp.toDIMACS() << " by comp at level 0");
			sstate.stats				->nbr_comp_bcp_unit++;
			if(!enqueue(imp,0))	{some_lits.clear(); return imp.flipped();}
		}
		some_lits.clear();
		sendUnits();
		return TRUE_LIT;
	}

	LOG2("Found " << some_lits.size() << " common lits at level " << dec_l);
	uint32_t levels					= 0; 		//will be set by "collectFalse"
	Literal latest_false 			= FALSE_LIT;//will be set by "collectFalse"
	vector<Literal> fls_lits;
	uint32_t reason_lts				= filterFalse(fls_lits,cls,levels,latest_false);

	asrt(latest_false != FALSE_LIT); 	asrt(sstate.vinf.litFalse(latest_false));
	asrt(!sstate.vinf.litFix(latest_false)); asrt(sstate.vinf.litAssLev(latest_false)==dec_l);

	if(reason_lts == max32){
		LOG2("Hyper Hyper Binary Resolution during Comp BCP creating "<< some_lits.size() << " binaries");
		while(!some_lits.empty()){
			Literal imp 		= some_lits.back(); some_lits.pop_back();
			if(sstate.vinf.litKnwn(imp)) continue;

			LOG3("Enqueue lit " << imp.toDIMACS() << " by comp at level " << dec_l<< " with bin reason " << latest_false.toDIMACS());
			sstate.addBinAndSend(imp,latest_false);
			sstate.stats->nbr_comp_bcp_bin++;
			if(!enqueue(imp,dec_l,latest_false)){some_lits.clear(); return imp.flipped();}
		}
		asrt(some_lits.empty()); asrt(openBCPImpls());
		return TRUE_LIT;
	}

	LOG2("Comp BCP implying "<< some_lits.size() << " variables"); asrt(lit_flags.allCleared());

	for(uint32_t t=ukwn.size();t--;){setFlag(ukwn[t]);} asrt(reason_lts>1);	asrt(fls_lits.size()==reason_lts);
	fls_lits.push_back(TRUE_LIT);

	while(!some_lits.empty()){
		Literal imp 		= some_lits.back(); some_lits.pop_back();
		if(sstate.vinf.litKnwn(imp)) continue;
		sstate.stats				->nbr_comp_bcp_long++;
		fls_lits[reason_lts]= imp;

		bool subsuming		= isSet(imp); asrt(cls->getUsers().hasAnyUser());

		SharedClause* ncls	= SharedClause::construct(sstate.s_info->all_users_bits,fls_lits.begin(),fls_lits.end(),levels,
							 !subsuming ||  cls->isLearnt() );


		LOG2("Create new clause " << (uint64_t)ncls << " for comp bcp " );

		ClsPointer clp 		= sstate.linkCls(ncls,fls_lits.size(),imp,latest_false, cls->getUsers()); sanity(asrtClsUnit(ncls,imp	));

		//if(subsuming){		cls->setWaste("Selfsubsume send new size ", fls_lits.size()); sstate.stats ->nbr_comp_selfsub++;}
		if(subsuming){		cls->setWaste(); sstate.stats ->nbr_comp_selfsub++;}

		LOG3("Enqueue lit " << imp.toDIMACS() << " by comp at level " << dec_l<< " with lbd " << levels);
		if(!enqueue(imp,dec_l,clp,ncls->getOrigLBD())){some_lits.clear();zeroFlaggs(ukwn); return imp.flipped();}

	}
	zeroFlaggs(ukwn);
	asrt(some_lits.empty()); asrt(openBCPImpls()); asrt(lit_flags.allCleared());
	return TRUE_LIT;
}

BCPETEMP
int32_t	BCPECLASS::bcpBeyondUP(){

	TIME_ON(sstate.stats->bcp_comp_time); asrt(!openBCPImpls());
	int32_t res = BCP_SUCCESS;

	while(beyUP_head < beyUP_cands.size()){
		LOG3("BCP Comp with head " << beyUP_head << " and comp clses " << beyUP_cands.size());

		ClsWrp& cw						= sstate.cinf.get(beyUP_cands[beyUP_head++]);
		if(!cw.shallDoBUP())			continue;
		cw								.setBUP(false); //mark that this clause was used for comp propagation

		Literal bcp_comp				= bcpBeyondUP(cw.getCls());

		if(bcp_comp==TRUE_LIT){			TIME_OFF(sstate.stats->bcp_comp_time); asrt(openBCPImpls()); return BCP_SUCCESS;}
		if(bcp_comp!=INVALID_LIT){		res=  bcpInitConflictUnit(bcp_comp);break;}
	}

	TIME_OFF(sstate.stats->bcp_comp_time);
	beyUP_cands	.clear();
	beyUP_head 	= 0;
	return res;
}



BCPETEMP
int32_t	BCPECLASS::bcpInitConflict(ClsPointer clp){
	asrt(lit_flags.allCleared());
	asrt(level_flags.allCleared());
	confl_levs 		= 0;
	conflicting		= clp;
	lemma			.clear();
	confl_lits		.clear();

	return 0;
}

BCPETEMP
int32_t	BCPECLASS::bcpInitConflictUnit	(const Literal& lit){
	asrt(lit_flags.allCleared());
	asrt(level_flags.allCleared());
	confl_levs 		= 0;
	conflicting		= ClsPointer();
	confl_lits		.clear();
	lemma			.clear();
	lemma			.push_back(lit);
	LOG2("Init conflict with " << lit.toDIMACS());
	return 1;
}




//helper function
//go back the implication graph
BCPETEMP
int32_t BCPECLASS::markConflictLit(const Literal& false_lit, const int32_t& dec_level){

	LOG3("markConflictLit " << false_lit.toDIMACS() << " at lev " << dec_level);
	if(isSet(false_lit)) return 0;
	asrt(sstate.vinf.litFalse(false_lit));

   int level = sstate.vinf.litAssLev(false_lit);

   if(level == 0){ LOG3(" skip clause lev " << level); return 0;}

   decision->litInConflict(false_lit);
   setFlag(false_lit);

   if(level <  dec_level){
	   lemma	.push_back(false_lit);
	   setLevel(level);
	   if(bump_glue){confl_lits.push_back(false_lit);}
	   return 0;
   }

   return 1;
}

//helper function for self subsumption
//resturns true if the given literal is necessary for the conflict clause and search
//cannot continue recursively with the assign reason for this lit
BCPETEMP
bool BCPECLASS::needLitGraphCheckHelper(const Literal& lit){

	if(isSet(lit)) return false;
	Variable var = lit.toVar();

	const uint32_t level = sstate.vinf.varAssignLevel(var);
	if(level > 0){
		if(sstate.vinf.varIsDecision(var) || !levelSet(level)){
			return true;
		}

		setFlag(lit);
		tmp_literal_store.push_back(lit);
	}
	return false;
}



BCPETEMP
bool 	BCPECLASS::reserveForVariables(uint32_t need_sz){
	if(need_sz <= sstate.stats->nbr_vars) return false;

	sstate.stats->nbr_vars = need_sz;
	//Variable infos
	sstate.vinf			.ensureSize(need_sz);
	sstate.cls_reg		.ensureVars(need_sz);
	decision			->ensureSize(need_sz);

	//Literal infos
	need_sz 		= need_sz << 1;
	sstate.wtch    	.enlarge(need_sz);
	lit_flags		.ensureSize(need_sz,true);
	return true;
}

BCPETEMP
void	BCPECLASS::reserveForClauses(uint32_t need_sz){
	sstate.cinf		.allocate(need_sz);
	sstate.cls_reg.reserveClses(need_sz);
}

BCPETEMP
void	BCPECLASS::sendUnits(){
	uint32_t to 	= decisionsAtLevel0();
	while(units_sent < to){
		sstate.s_info->sendUnitCls(assign_stack[units_sent++]);}

}

BCPETEMP
bool	BCPECLASS::asrtAllUnitsSent() const{
	return decisionsAtLevel0() == units_sent;
}

BCPETEMP
bool	BCPECLASS::asrtNoReplNoFixLits(SharedClause* cls){
	bool ok = true;

	FOR_LITS_STRT(cls,lit);
		if(sstate.vinf.litFix(lit) || sstate.vinf.litRepl(lit)){
			ok = false;
			ASRT_PRNT("Lit " <<lit.toDIMACS() << " failed in asrtNoReplNoFixLits");}

	FOR_LITS_END;

	return ok;
}




BCPETEMP
bool	BCPECLASS::asrtContaisRepl(SharedClause* cls){
	FOR_LITS_STRT(cls,lit);
		if(sstate.vinf.litRepl(lit)){return true;}
	FOR_LITS_END;

	FOR_LITS_STRT(cls,lit);
		cout << lit.toDIMACS() << " ";
	FOR_LITS_END;
	cout << endl;

	return false;
}



BCPETEMP
Literal		BCPECLASS::importBinImpliesUnit(Literal a, const Literal& b){

	//sstate.cls_reg.sneakPreview(a,b);
	if(	sstate.vinf.litFixTrue(a) || sstate.vinf.litFixTrue(b) || a == b.flipped() ||
			sstate.vinf.varElim(a.toVar()) || sstate.vinf.varElim(b.toVar()))return TRUE_LIT;

	if(a==b || sstate.vinf.litFixFalse(a))	return b;
	if(sstate.vinf.litFixFalse(b)) 			return a;

	asrt(!sstate.vinf.litFix(a)); asrt(!sstate.vinf.litFix(b));
	if(!sstate.vinf.litFalse(a) && !sstate.vinf.litFalse(b)) {return INVALID_LIT;}

	LOG2("BCP import "<<a.toDIMACS() << ", " << b.toDIMACS());

	if(sstate.vinf.litFalse(a)){
		uint32_t a_lev = sstate.vinf.litAssLev(a); asrt(sstate.vinf.litAssLev(sstate.vinf.litRoot(a))==a_lev);

		if(sstate.vinf.litKnwn(b)){
			uint32_t b_lev = sstate.vinf.litAssLev(b);
			if(sstate.vinf.litFalse(b)){ asrt(sstate.vinf.litAssLev(sstate.vinf.litRoot(b))==b_lev);
				if(a_lev==b_lev){asrt(a_lev>0);
					backjumpKeepLevel(a_lev-1); asrt(sstate.vinf.litUkwn(a)); asrt(sstate.vinf.litUkwn(b));
					return INVALID_LIT;}

				if(a_lev > b_lev){
					backjumpKeepLevel(b_lev); asrt(sstate.vinf.litUkwn(a));
					return enqueue(a,b_lev,b) ? INVALID_LIT : a.flipped();}

			}else{asrt(sstate.vinf.litTrue(b));
				if(b_lev <= a_lev) {return INVALID_LIT;}
			}
		}

		backjumpKeepLevel(a_lev); asrt(sstate.vinf.litUkwn(b)); asrt(sstate.vinf.litFalse(a));
		return enqueue(b,a_lev,a) ? INVALID_LIT : b.flipped();
	}

	asrt(sstate.vinf.litFalse(b)); asrt(!sstate.vinf.litFalse(a));
	uint32_t b_lev = sstate.vinf.litAssLev(b); asrt(sstate.vinf.litAssLev(sstate.vinf.litRoot(b))==b_lev);
	if(sstate.vinf.litTrue(a) && sstate.vinf.litAssLev(a) <= b_lev) { return INVALID_LIT;}

	backjumpKeepLevel(b_lev);
	return (sstate.vinf.litTrue(a) || enqueue(a,b_lev,b) )?INVALID_LIT:a.flipped();
}


BCPETEMP
bool BCPECLASS::imprtLong(SharedClause* orig,uint32_t max_lbd, uint32_t balance_percent, act_t threash){
	asrt(asrtAllUnitsSent());
	uint32_t quality		= 100;
	SharedClause* n_cls 	= orig;
	uint32_t nsz			= 0;
	if(orig->isWaste() || orig->getOrigLBD()>max_lbd ){asrt(orig->isLearnt() || orig->isWaste());
		quality 	= max32;
	}else{
		if(orig->isLearnt()){quality=balance_percent;}
		n_cls = createReplacedIf(orig,nsz,quality,threash);
		if(n_cls == NULL)	{asrt(!ok()); return false;}
	}

	if(quality==max32 || n_cls->isWaste()){asrt(n_cls==orig);
		SharedClause::release(n_cls,sstate.s_info->user_bit);return true;}
		//SharedClause::release(n_cls,sstate.s_info->user_bit,"not good enough at imprtLong");return true;}

	asrt(asrtNoReplNoFixLits(n_cls));
	Literal w1,w2;
	findWatchers(n_cls,w1,w2);  asrt(!sstate.vinf.litFalse(w2) || decisionLevel()!= 0);

	const bool chng	= n_cls != orig; asrt(nsz!=0);
	ClsPointer clp = sstate.linkCls(n_cls,nsz,w1,w2, chng?orig->getUsers():UserMask(0),sstate.s_info->getImportLbdMsk());
	if(chng){
		//orig->setWaste("Import long created new one with " , nsz);
		//if(orig->isFix()){LOG0("at level " << decisionLevel() << " new one is: "); n_cls->print();}
		//SharedClause::release(orig,sstate.s_info->user_bit,"replaced by better in longImprt");

		orig->setWaste();
		SharedClause::release(orig,sstate.s_info->user_bit);
	}

	if(sstate.vinf.litTrue(w1) || !sstate.vinf.litFalse(w2)) return true;

	LOG3("Found clause to enqueue" << w1.toDIMACS());
	asrt(asrtClsUnit(n_cls,w1)); asrt(decisionLevel()!=0);

	if(!enqueue(w1,decisionLevel(),clp,sstate.cinf.get(clp).getLBD())){
		LOG1("Jump back in cls import after enqueue" << w1.toDIMACS());
		backjumpKeepLevel(0);
		return enqueue(w1.flipped(),0) && (bcpLong() == BCP_SUCCESS);}
	return true;
}



BCPETEMP
void BCPECLASS::applyVarElim(const Variable& var,vector<Literal>& lits_to_update){
	asrt(decisionLevel()==0);
	LOG1("Apply Eliminat var " << var.toDIMACS());
	if(sstate.vinf.varKnwn(var)) return;

	sstate.vinf		.setElim(var);
	decision	->forgetVar(var); //may appear again since clauses with var are only deleted lazyly
	sstate.stats->nbr_var_elim++;

	wasteAllWatched(var.toPosLit());
	wasteAllWatched(var.toNegLit());

	sstate.wtch.removeVar(var,lits_to_update);
}



BCPETEMP
bool BCPECLASS::replaceVar(const Literal& die, const Literal& survive,vector<Literal>& units_out,
		vector<Literal>& update_bins,vector<ClsPointer>& update_cls){

	LOG3("Replace lit " << die.toDIMACS() << " by " << survive.toDIMACS());
	bool new_repl = false;
	Literal df = die.flipped(); asrt(!sstate.vinf.litRepl(survive));
	if(!sstate.vinf.litRepl(die)){ asrt(!sstate.vinf.litRepl(die.flipped()));
		new_repl = true;
		update_bins.push_back(survive);
		update_bins.push_back(survive.flipped());

		sstate.wtch.addBins(die,survive,update_bins,units_out);
		sstate.wtch.addBins(df,survive.flipped(),update_bins,units_out);

		sstate.wtch.addClsesOf(die,update_cls);
		sstate.wtch.addClsesOf(df,update_cls);

		sstate.wtch.clearOf(die);
		sstate.wtch.clearOf(df);

		sstate.vinf.setRepl(die,survive);
	}

	asrt(sstate.vinf.litRepl(die)); asrt(sstate.vinf.litRepl(die.flipped()));
	return new_repl;
}


BCPETEMP
bool BCPECLASS::applyLitReplsAndRemoves(const vector<Message>& repls_removes){
	LOG1("applyLitReplsAndRemoves with messages: " << repls_removes.size());

	backjumpKeepLevel(0);
	if(bcpLong() != BCP_SUCCESS) return setInvalid("Conflict before lit replace");
	sendUnits();

	sanity(sstate.wtch.asrtBinsBothDirs(sstate.vinf));
	Literal lit, repl;
	vector<Literal> units; vector<Literal> update_lits; vector<ClsPointer> update_clses;
	for(uint32_t i =0; i != repls_removes.size();++i){
		const Message& m = repls_removes[i]; asrt(m.isLitRepl() || m.isVarElim());
		if(m.isVarElim()){
			applyVarElim(m.getVarElim(),update_lits);
			continue;
		}
		lit = m.getLitRepl(repl);
		if(sstate.vinf.litFix(lit)) continue;
		asrt(!sstate.vinf.litFix(repl));
		sstate.stats->nbr_var_repl += replaceVar(lit,repl,units,update_lits,update_clses);
		sstate.cls_reg.delOccs(lit.toVar());

		//experimet mit replVar
		//decision->forgetVar(lit.toVar());
	}


	updateReplsInBins(update_lits,units);


	for(uint32_t i=units.size();i--;){
		lit = units[i];
		if(sstate.vinf.litTrue(lit)) continue;
		if(sstate.vinf.litFalse(lit) || !enqueue(lit,0)){
			return setInvalid("Contradiction in Lit replace");}
	}
	sendUnits();

	uint32_t quality_dummy; const act_t border_dummy(0); uint32_t nsz (0);
	uint32_t cp = 0; //keep those clauses that have to be freed
	for(uint32_t i=0;i != update_clses.size();++i){
		ClsPointer clp		= update_clses[i];
		ClsWrp& cw			= sstate.cinf.get(clp);
		if(cw.watchersInvalid()){//seen twice now -> none of both watchers will see it in clearWatches
			update_clses[cp++] = clp; continue;} //free later on, such that no freed pointers are reused immediately (for cleaning occurrence lists)

		SharedClause* orig = cw.getCls();
		cw.invalidateWatchers(); //mark for deletion for other watcher

		if(orig->isWaste()) continue; asrt(asrtAllUnitsSent());

		SharedClause* ncls = createReplacedIf(orig, nsz, quality_dummy=100,border_dummy); asrt(quality_dummy!=max32);
		if(ncls == NULL){	return setInvalid("Bad lit replace");} asrt(orig->isWaste() || ncls!=orig);
		if(ncls != orig){
			Literal w0 = ncls->getAt(0);
			Literal w1 = ncls->getAt(1);		asrt(sstate.vinf.litUkwn(w0)); asrt(sstate.vinf.litUkwn(w1)); asrt(orig->getUsers().hasAnyUser());
			sstate.linkCls(ncls,nsz,w0,w1, orig->getUsers());
			orig->setWaste();
			//orig->setWaste("Replace if got new clause size ", nsz);
		}

	}


	for(ClsPointer iter = sstate.cinf.lastValid();!iter.isNull(); iter = sstate.cinf.nextValidBefore(iter)){
		LOG3("applyLitRepls " << iter.idx()); asrt(asrtAllUnitsSent());

		SharedClause* orig 	= sstate.cinf.get(iter).getCls();
		SharedClause* ncls	= createReplacedIf(orig,nsz,quality_dummy=100,border_dummy); asrt(quality_dummy!=max32);

		if(ncls == NULL){ asrt(!ok()); return false;}
		if(ncls != orig){
			Literal w0 = ncls->getAt(0);
			Literal w1 = ncls->getAt(1);		asrt(sstate.vinf.litUkwn(w0)); asrt(sstate.vinf.litUkwn(w1)); asrt(orig->getUsers().hasAnyUser());
			sstate.linkCls(ncls,nsz,w0,w1,orig->getUsers());
			orig->setWaste();
		}
	}

	//free clauses now
	while(cp--){sstate.cinf.freeCls(update_clses[cp],sstate.s_info->user_bit);}
	clearWatches(max32);

	bool ret = bcpLong() == BCP_SUCCESS;
	sendUnits(); sanity(sstate.wtch.asrtBinsBothDirs(sstate.vinf));
	LOG1("applyLitReplsAndRemoves finished with : " << repls_removes.size());
	return ret;
}


/**make sure that all units are sent before**/
BCPETEMP
SharedClause* BCPECLASS::createReplacedIf(SharedClause* cls,uint32_t& sz_out,uint32_t& good_percent,act_t border){
	if(cls->isWaste()) return cls;

	vector<Literal> nw_lits;
	asrt(nw_lits.empty());

	bool chng = false; uint32_t good = 0;asrt(asrtAllUnitsSent());

	FOR_LITS_STRT(cls,lit);
		lit = sstate.vinf.litRepl(lit) ? (chng=true,sstate.vinf.findLitRepl(lit)) : lit;

		if(sstate.vinf.litFixTrue(lit)||sstate.vinf.varElim(lit.toVar())){ asrt(asrtAllUnitsSent());
			cls->setWaste();return cls;}

		if(sstate.vinf.litFix(lit)){ asrt(sstate.vinf.litFalse(lit)); chng=true; continue;}

		nw_lits.push_back(lit);

		//better take activity than number of false lits
		good				+= (decision->getAct(lit) >= border);
		//sstate.cls_reg		.sneakPreview(cls,lit,max_ungood_lits);

	FOR_LITS_END;

	sz_out = nw_lits.size();
	if(good_percent*sz_out < good * 100){
		good_percent = max32; asrt(cls->isLearnt());
		return cls;
	}

	if(!chng) return cls;

	std::sort(nw_lits.begin(),nw_lits.end());

	sz_out = 0; Literal prev = INVALID_LIT;
	for(uint32_t i=0;i!=nw_lits.size();++i){
		Literal lit = nw_lits[i];
		if(prev == lit) continue;
		if(prev == lit.flipped()){cls->setWaste(); return cls;}
		nw_lits[sz_out++] = prev = lit;
	}

	if(sz_out < 3){

		if(sz_out == 2){
			if(!sstate.wtch.containsBin(nw_lits[0],nw_lits[1])){
				sstate.addBinAndSend(nw_lits[0],nw_lits[1]);}
		}else{
			backjumpKeepLevel(0); asrt(sz_out==0 || sstate.vinf.litUkwn(nw_lits[0]) );
			if(sz_out==0 || !enqueue(nw_lits[0],0)){
				setInvalid("createReplaced produced conflict");
				return NULL;
			}
		}

		LOG2("Set Waste clause " << ((uint64_t)&cls) << " replaced by " << sz_out);
		sendUnits();
		//cls->setWaste("Subsumed by cls of size ",sz_out);
		cls->setWaste();
		return 	cls;
	}

	LOG3("shrink clause to size " << sz_out); asrt(nw_lits[0]!=nw_lits[1]);
	//check new long clause
	return SharedClause::construct(sstate.s_info->all_users_bits,nw_lits.begin(),nw_lits.begin()+sz_out,
								min(cls->getOrigLBD(),sz_out),cls->isLearnt());
}


BCPETEMP
void BCPECLASS::updateReplsInBins(vector<Literal>& lits,vector<Literal>& units_out){
	asrt(decisionLevel()==0); asrt(lit_flags.allCleared()); asrt(sstate.vinf.asrtAllFlagsClear());

	LOG3("START updateReplsInImpls with vec sz " << lits.size());
	Literal other;
	for(uint32_t i=lits.size();i--;){
		Literal lit = lits[i]; LOG3("clear impls of lit " << lit.toDIMACS());

		if(sstate.vinf.isFlagged(lit) || sstate.vinf.litRepl(lit)) continue; //already cleaned
		sstate.vinf.setFlag(lit); asrt(!sstate.vinf.litRepl(lit));

		WIter<Literal> iter = sstate.wtch.getWBinsOf(lit);

		if(!sstate.vinf.varElim(lit.toVar())){
			setFlag(lit.flipped());

			while(iter.getNext(other)){
				other = sstate.vinf.findLitRepl(other); asrt(!sstate.vinf.litRepl(other));
				if(isSet(other) || sstate.vinf.varElim(other.toVar())) continue;
				setFlag(other);
				iter.keep(other);
				if(isSet(other.flipped())){
					LOG2("Unit implied ad impl update "<< lit.toDIMACS());
					units_out.push_back(lit);}
			}
		}

		sstate.wtch.finish(lit,iter);

		zeroFlags(lit.flipped());
		Iter<Literal> it = sstate.wtch.getBinsOf(lit);
		while(it.getNext(other)){zeroFlags(other);}
	}



	for(uint32_t i=lits.size();i--;){sstate.vinf.clearFlag(lits[i]);}
	asrt(lit_flags.allCleared()); asrt(sstate.vinf.asrtAllFlagsClear());
}


BCPETEMP
bool BCPECLASS::import(uint32_t max_lbd,uint32_t balance_percent){
	sstate.stats->nbr_imports++;
	sendUnits(); asrt(max_lbd >=2);
	Message m; Literal a,b; UserMask recommended(0);
	TIME_ON(sstate.stats->import_long_time);
	vector<Message> repls_removes;

	typename decision_t::act_t max_hf = decision->getHalfOfMax();

	while(true){

		while(sstate.s_info->popMessage(m)){
			LOG2("Import Message ");

			if(m.isBinCls()){

				m.getBinCls(a,b);
				a = sstate.vinf.findLitRepl(a); b = sstate.vinf.findLitRepl(b);
				LOG3("Received bin " << a.toDIMACS() << " " << b.toDIMACS() );

				Literal unit = importBinImpliesUnit(a,b);
				if(unit == TRUE_LIT){ 			continue;}
				if(unit == INVALID_LIT)	{ asrt(!sstate.vinf.litFalse(a) || !sstate.vinf.litFalse(b));
					sstate.cls_reg.registerBin(a,b);
					sstate.wtch.addBinClsIf(a,b); 	continue;}

				m = Message::newUnitCls(unit);
			}

			if(m.isUnitCls()){
				a = sstate.vinf.findLitRepl(m.getUnitCls());
				if(sstate.vinf.litFixTrue(a)) 		continue;
				LOG2(" Jump back to 0 for unit import " << a.toDIMACS());

				backjumpKeepLevel(0);
				sstate.vinf.unsetElim(a.toVar());
				if(sstate.vinf.litFalse(a) || !enqueue(a,0) || (bcpLong() != BCP_SUCCESS)){
					return setInvalid("Import unit");}

				sendUnits();

			}else if(m.isLongCls()){

				SharedClause* cls 	= m.getLongCls(recommended); asrt(asrtAllUnitsSent());
				const bool should	= recommended.hasUser(sstate.s_info->user_bit); //LOG0("Receive long at " << ((uint64_t)cls) << " is waste is " << cls->isWaste() << " fix " << cls->isFix() << " should is " << should);

				if(!imprtLong(cls,should?max32:max_lbd,should?100:balance_percent,max_hf)){
					LOG1("Clause " << ((uint64_t)cls) << " is unsat"); TIME_OFF(sstate.stats->import_long_time);
					return setInvalid("Import Longs");
				}
			}else if(m.isBinRemove()){

				m.getBinRemove(a,b); //do it immediately even if (a,b) is reason to an assignment
				if(sstate.varReplFixOrElim(a.toVar()) || sstate.varReplFixOrElim(b.toVar())) continue;

				TwiceVisitor visi(sstate.vinf,b);
				BFS(a.flipped(),visi,true,true);
				if(visi.cnt < 2){ continue;}

				sstate.wtch.removeBinPart(a,b);
				sstate.wtch.removeBinPart(b,a);
				sstate.stats->nbr_scc_shortcuts++;


			}else{
				asrt( m.isLitRepl() || m.isVarElim());
				repls_removes.push_back(m);
			}
		}

		sendUnits();
		if(repls_removes.empty()) 	return true;

		if(!repls_removes.empty()){
			if(!applyLitReplsAndRemoves(repls_removes)) return setInvalid("applyLitRepls");
			repls_removes.clear(); asrt(asrtAllUnitsSent());
		}
	}
	TIME_OFF(sstate.stats->import_long_time);

	return true;
}



BCPETEMP
void BCPECLASS::findWatchers(SharedClause* cls, Literal& w1, Literal& w2){
	Literal sec_fls = FALSE_LIT; w1=w2 =INVALID_LIT; asrt(asrtNoReplNoFixLits(cls));
	uint32_t max_fls=0; uint32_t max_fls2=0; uint32_t min_tru=max32;

	FOR_LITS_STRT(cls,lit);
		asrt(!sstate.vinf.litFix(lit)); asrt(!sstate.vinf.litRepl(lit));

		if(!sstate.vinf.litFalse(lit)){
			if(w1!=INVALID_LIT){	w2=lit;asrt(w1!=w2); return;}

			w1						= lit;	asrt(min_tru==max32);
			if(sstate.vinf.litUkwn(w1)){	continue;}

			min_tru 				= sstate.vinf.litAssLev(w1); asrt(min_tru > 0);
		}else{
			uint32_t lev 			= sstate.vinf.litAssLev(lit);
			if(lev <= max_fls){
				if(lev>=max_fls2)	{sec_fls=lit;max_fls2=lev;}	continue;}

				max_fls2	= max_fls;
				max_fls  	= lev;
				sec_fls 	= w2;
				w2 			= lit;
		}

		if(min_tru <= max_fls){
			asrt(w1!=INVALID_LIT); asrt(min_tru>0); asrt(w2!=INVALID_LIT); asrt(w1!=w2);
			asrt(!sstate.vinf.litFalse(w1) || !sstate.vinf.litFalse(w2));
			return;}

	FOR_LITS_END;


	if(w1==INVALID_LIT){ //only false literals
		asrt(max_fls > 0);asrt(max_fls!=max_fls2 || max_fls2!=0); asrt(max_fls2 <= max_fls);

		max_fls = max_fls2 - (max_fls==max_fls2);
		LOG3("have to backjump to level "<<max_fls);

		w1 = w2;
		w2 = sec_fls;
	}

	asrt(w1 != INVALID_LIT); asrt(w2 != INVALID_LIT); asrt(w1 != w2);

	backjumpKeepLevel(max_fls); asrt(min_tru>max_fls); asrt(sstate.vinf.litUkwn(w1));
	asrt(sstate.vinf.litUkwn(w1)); asrt(sstate.vinf.litUkwn(w2) || asrtClsUnit(cls,w1));
}








BCPETEMP
Literal BCPECLASS::checkSatByLit(const SharedClause* cls) const{
	FOR_LITS_STRT(cls,cand);
		if(sstate.vinf.litTrue(cand)) return cand;
	FOR_LITS_END;

    return INVALID_LIT;
}


BCPETEMP
void BCPECLASS::backjumpZero(){

	LOG3("Jump back to level zero ");
	if(decisionLevel() == 0) return;

	uint32_t assign_idx = decision_level_end[0];
	asrt(assign_stack.size() >= assign_idx);

	queue_begin 		= assign_stack.size();
	Variable x 			= INVALID_VAR;
	Literal l			= INVALID_LIT;
	while(queue_begin != assign_idx ){
		l	= assign_stack[--queue_begin];
		x  	= l.toVar();
		sstate.vinf		.varClearValue(x);
		decision	->restoreVar(x);
		zeroFlags	(l);
	}

	assign_stack		.resize(queue_begin);
	beyUP_cands			.clear();
	beyUP_head 			= 0;
	decision_level_end	.clear();
}




BCPETEMP
void BCPECLASS::backjumpKeepLevel(uint32_t level){//,const bool flip){

	LOG3("Jump back to level " << level << " from " << decisionLevel());
	if(decisionLevel() <= level) return;
	TIME_ON(sstate.stats->bcktrk_time);

	Variable x 			= INVALID_VAR;
	uint32_t assign_idx = decision_level_end[level];
	asrt(assign_stack.size() >= assign_idx);

	queue_begin 	= assign_stack.size();

	while(queue_begin != assign_idx ){
		x  = assign_stack[--queue_begin].toVar();
		sstate.vinf		.varClearValue(x);
		decision	->restoreVar(x);}


	assign_stack		.resize(queue_begin);
	beyUP_cands			.clear();
	beyUP_head 			= 0;

	sstate.stats->sum_bcktrk += decisionLevel() - level;
	decision_level_end.resize(level);
	TIME_OFF(sstate.stats->bcktrk_time);
}



BCPETEMP
void	BCPECLASS::newDecisionLevel(){
	asrt(!openBCPImpls()); asrt(!openCompImpls());
	decision_level_end.push_back(assign_stack.size());
	if(decision_level_end.size() >= sstate.stats->max_dec_level){
		sanity(level_flags.allCleared());
		level_flags.ensureSize(decision_level_end.capacity()+1,false);
		sstate.stats->max_dec_level = decision_level_end.size();
	}
	asrt(beyUP_cands.empty()); asrt(beyUP_head==0);
}




//boolean constraint propagation
BCPETEMP
Literal	BCPECLASS::enqueueBinBCP(const Literal& lit,const Literal& root){

	TIME_ON(sstate.stats->bcp_time); asrt(root == sstate.vinf.litRoot(lit.flipped())); asrt(!sstate.vinf.litRepl(lit));
	const uint32_t dec_level 	= decisionLevel();
	uint32_t next				= assign_stack.size(); asrt(sstate.vinf.litFalse(root)); asrt(sstate.vinf.litAssLev(root)==dec_level);
	assign_stack				.push_back(lit);

	Literal other;

	while(next != assign_stack.size()){
		//++sstate.stats->nbr_bcp;
		const Literal true_lit		= assign_stack[next++];
		const Literal false_lit 	= true_lit.flipped();
		Iter<Literal> bi			= sstate.wtch.getBinsOf(false_lit);

		LOG3("Propagate lit " << true_lit.toDIMACS() << " at lev " << dec_level);

		while(bi.getNext(other)){

			LOG3("check literal " << other.toDIMACS() << " at lev " << dec_level);
			if(sstate.vinf.litTrue(other)) continue;

			if(sstate.vinf.litFalse(other)){
				asrt(sstate.vinf.litAssLev(other)==dec_level); TIME_OFF(sstate.stats->bcp_time);
				LOG3("Bin false " << other.toDIMACS());
				return other;}

			asrt(sstate.vinf.litUkwn(other)); //sanity(sstate.wtch.asrtPathExists(root.flipped(),other)); sanity(sstate.wtch.asrtPathExists(other.flipped(),root));
			LOG3("Bin set value " << other.toDIMACS());
			sstate.vinf.setValueReason(other,dec_level,false_lit,root);
			assign_stack.push_back(other);
		}
		LOG3("Finish propagate lit " << true_lit.toDIMACS() << " at lev " << dec_level);

	}
	LOG3("Finish propagation with stack " << next);
	TIME_OFF(sstate.stats->bcp_time);
	return INVALID_LIT;
}

BCPETEMP
void BCPECLASS::bumpGlueIf(){
	if(!bump_glue || confl_levs >= SharedClause::LBD_MAX){ return;}

	for(uint32_t i=confl_lits.size();i--;){
		const BDReason& ri = sstate.vinf.varAssignReason(confl_lits[i].toVar()); asrt(sstate.vinf.litKnwn(confl_lits[i]));//asrt(ri.isClsReason());

		if(ri.getLBD() < confl_levs){
			asrt(ri.isClsReason());aux_var(ClsWrp cw = sstate.cinf.get(ri.getClsReason());)asrt(cw.getLBD() < confl_levs);
			decision->litInConflict(confl_lits[i]);
		}
	}
}

BCPETEMP
bool 	BCPECLASS::handleConflict(int32_t bcp_res){

	if(decisionLevel() == 0) {asrt(bcp_res==0 || bcp_res==1);
		return setInvalid("Conflict at level 0"); }

	LOG3("Start FUIP ");
	bool lemma_in_db	= firstUIP(bcp_res); asrt(lemma.size()>0);
	Literal watcher1	= lemma.back();	asrt(watcher1 != INVALID_LIT); asrt(sstate.vinf.litAssLev(watcher1) == decisionLevel());

	bumpGlueIf();

	LOG3("Start Lemma Reduction with size: "<< lemma.size() <<" lemma_in_db " << lemma_in_db );
	lemmaReduction(lemma_in_db);

	LOG3("Determine Backtrack level with size: "<< lemma.size() <<" lemma_in_db " << lemma_in_db );
	Literal watcher2;
	uint32_t backtrack_level = backtrackLevel(watcher2);


	/*uint32_t dist = decisionLevel()-backtrack_level;
	if(dist-- > 1){
		Literal aut_guess[dist];
		uint32_t dec = backtrack_level;
		for(uint32_t a=0;a!=dist;++a){ asrt(dec< decisionLevel());
			aut_guess[a] = assign_stack[decision_level_end[dec++]]; //decision dec+1
		}

		///
		uint32_t first	= decision_level_end[backtrack_level];
		uint32_t end 	= decision_level_end.back(); asrt(first<end);
		dist			= end-first;
		Literal aut_guess[dist]; uint32_t a=0;
		while(first!=end){aut_guess[a++]=assign_stack[first++];}///

		sstate.s_info->pushAutarkyGuess(aut_guess,dist);
	}
	 */

	LOG3("Finally Backtrack to level " << backtrack_level);asrt(backtrack_level < decisionLevel());
	backjumpKeepLevel(backtrack_level);		asrt(sstate.vinf.litUkwn(watcher1) && decisionLevel() == backtrack_level);

	LOG3("Learn lemma ");
	if(watcher2 == FALSE_LIT){
		LOG2("Resolved fixed literal: "<<watcher1.toDIMACS()); asrt(backtrack_level == 0);
		return enqueue(watcher1,0);
	}

	asrt(decisionLevel()>0);

	decision->decay();

	//add lemma
	if(currLemmaLong()){ asrt(watcher2 != FALSE_LIT); asrt(!lemma_in_db || !conflicting.isNull());
		SharedClause* cls;

		if(!lemma_in_db){
			asrt(lemma.size()> 2); asrt(confl_levs>0); asrt(sstate.vinf.litUkwn(watcher1)); asrt(sstate.vinf.litFalse(watcher2));


			if(confl_levs < 5) { //sstate.s_info->getMaxImportLBD()){
				cls	= SharedClause::construct(sstate.s_info->all_users_bits,lemma.begin(),lemma.end(),confl_levs,true);

				LOG2("Create and shared new clause " << (uint64_t)cls << " as lemma" );	sanity(asrtClsUnit(cls,watcher1));
				conflicting 		= sstate.linkCls(cls,lemma.size(),watcher1,watcher2, UserMask(sstate.s_info->user_bit));
			}else{
				cls	= SharedClause::construct(UserMask(sstate.s_info->user_bit),lemma.begin(),lemma.end(),confl_levs,true);

				LOG2("Create dont share clause " << (uint64_t)cls << " as lemma" );	sanity(asrtClsUnit(cls,watcher1));
				conflicting 		= sstate.linkCls(cls,lemma.size(),watcher1,watcher2, UserMask(0));
			}


		}else{
			cls = sstate.cinf.get(conflicting).getCls();
		}
		asrt(sstate.cinf.validPtr(conflicting));

		confl_levs = cls->getOrigLBD();
		if(enqueue(watcher1,backtrack_level,conflicting,confl_levs)) return true;

	}else{
		if(!lemma_in_db){
			LOG2("Bin not in DB: " << watcher1.toDIMACS()<< " " << watcher2.toDIMACS()); asrt(!sstate.wtch.containsBin(watcher1,watcher2));
			sstate.addBinAndSend(watcher1,watcher2);
			//conflicting = ClsPointer();
		}else{
			LOG1("Bin already in DB: " << watcher1.toDIMACS()<< " " << watcher2.toDIMACS());
			asrt(conflicting.isNull()); asrt(sstate.wtch.containsBin(watcher1,watcher2));
		}
		LOG3("Enqueue bin lit " << watcher1.toDIMACS() << " reason " << watcher2.toDIMACS()); asrt(!openBCPImpls());
		if(enqueue(watcher1,backtrack_level,watcher2)) return true;
	}

	LOG1("Enqueue not immediately ok");
	backjumpKeepLevel(0);
	return !sstate.vinf.varKnwn(watcher1.toVar()) && enqueue(watcher1.flipped(),0);
}


BCPETEMP
uint32_t 	BCPECLASS::strengthenClsSetConflicting(SharedClause* cls,Literal p_lit){
	asrt(cls != NULL); asrt(otf_sub==DoOnTheFlySub); //asrt(asrtClsSorted(cls));

	int32_t w1l = -1; int32_t w2l = -1;
	vector<Literal> lits;
	int32_t lev; Literal w1,w2;

	FOR_LITS_STRT(cls,lit);
		if(lit == p_lit || sstate.vinf.litFix(lit))continue;
		lits.push_back(lit);	asrt(sstate.vinf.litFalse(lit));
		lev = (int32_t)sstate.vinf.litAssLev(lit);
		if(lev > w1l){		w2l=w1l; w2=w1; w1l=lev; w1=lit;}
		else if(lev>w2l){	w2l=lev;w2=lit;}
	FOR_LITS_END;

	const uint32_t new_sz = lits.size();
	if(new_sz < 2){
		LOG1("strengthen clause creates unit" <<w1.toDIMACS() << " lemma size is " << lemma.size());
		sstate.s_info->sendUnitCls(w1); asrt(lemma.empty());
		return 0;
	}

	LOG2("Strengthen clause at level " << decisionLevel() << " to sz " << new_sz);

	if(new_sz==2){
		sstate.addBinAndSend(w1,w2);
		LOG2("Create bin clause " << w1.toDIMACS() << " " << w2.toDIMACS() << " while strengthening" );
		cls->setWaste();
		conflicting = ClsPointer();
		return 1; // to indicate that cls is in database but "conflicting" is null
	}

	SharedClause* new_cls = SharedClause::construct(sstate.s_info->all_users_bits,lits.begin(),lits.end(),cls->getOrigLBD(),cls->isLearnt());
	LOG2("Create new clause " << (uint64_t)new_cls << " while strengthening" ); asrt(cls->getUsers().hasAnyUser()); //asrt(asrtClsSorted(new_cls));
	conflicting = sstate.linkCls(new_cls,new_sz,w1,w2,cls->getUsers());
	cls->setWaste();
	return new_sz;
}



/**Analyse the conflict
 * nbr_pivot_elements states how many literals of the current decision level are marked*/
BCPETEMP
bool BCPECLASS::firstUIP(int32_t nbr_pivots){
	asrt(nbr_pivots >= 1);
	if(nbr_pivots==1){asrt(lemma.size()==1);asrt(conflicting.isNull()); return false;}

	const uint32_t dec_level	= decisionLevel(); 	asrt(!conflicting.isNull());
	uint32_t as_idx  			= assign_stack.size() ;
	Literal 	p_lit,fp_lit;


	//if prev_res_sz != 0 the previous resolvent is contained in db and has this size (prev_res_sz means binary)
	uint32_t	prev_res_sz	= (otf_sub == DoOnTheFlySub) ? lemma.size()+nbr_pivots : 0; asrt(otf_sub==NoOnTheFlySub || asrtResolventSize(prev_res_sz));

	while(true){

		//skip all unseen assignments
		//since there always exists a UIP the index cannot become negative
		while ((p_lit = assign_stack[--as_idx],fp_lit=p_lit.flipped(),!isSet(fp_lit))){
			asrt(as_idx < assign_stack.size());}



		if(--nbr_pivots == 0){
			asrt(sstate.vinf.litAssLev(fp_lit) == dec_level); asrt(fp_lit != INVALID_LIT);
			lemma.push_back(fp_lit);		asrt(asrtAllLitsFlagged(lemma));
			bool lemma_in_db	= prev_res_sz != 0; LOG3("prev_res_sz  was " << prev_res_sz << " in FUIP, lemma size " << lemma.size());
			asrt(!lemma_in_db || lemma.size() != 2 || (conflicting.isNull() && sstate.wtch.containsBin(lemma[0],lemma[1]))); asrt(otf_sub==DoOnTheFlySub||!lemma_in_db);
			return lemma_in_db;
		}

		clearFlag(fp_lit);

		const BDReason& reason = sstate.vinf.varAssignReason(p_lit.toVar());
		asrt(sstate.vinf.litAssLev(p_lit) == dec_level);	asrt(!reason.isDecision());

		if(reason.isBinReason()){
			nbr_pivots += markConflictLit(reason.getBinReason(),dec_level);
							asrt(asrtResolventSize(lemma.size()+nbr_pivots));

			if((otf_sub == DoOnTheFlySub) && (lemma.size() + nbr_pivots < prev_res_sz)){asrt(!conflicting.isNull());
					LOG2("Strengthen cls after bin reason");
					prev_res_sz = strengthenClsSetConflicting(sstate.cinf.get(conflicting).getCls(),fp_lit);
			}else{prev_res_sz= 0;}
			asrt(prev_res_sz<=1 || lemma.size()+nbr_pivots==prev_res_sz);

		}else{
			asrt(reason.isClsReason());
			SharedClause* cls	= sstate.cinf.get(reason.getClsReason()).getCls(); asrt(cls != NULL);
			uint32_t cls_rel	= 0;
			cls->incConflCnt();

			FOR_LITS_STRT(cls,m_lit);
				if(m_lit ==p_lit) continue;
				nbr_pivots += markConflictLit(m_lit,dec_level);
				cls_rel	+= (uint32_t)(!sstate.vinf.litFix(m_lit));
			FOR_LITS_END;

			if(otf_sub == DoOnTheFlySub){
				uint32_t resolv_sz 	= lemma.size() + nbr_pivots; asrt(asrtResolventSize(resolv_sz)); asrt(resolv_sz>=cls_rel);
				bool stren_cls		=  resolv_sz == cls_rel;
				if(resolv_sz < prev_res_sz){asrt(!conflicting.isNull()); asrt(resolv_sz != 0);
					SharedClause* prev = sstate.cinf.get(conflicting).getCls();
					if(!stren_cls  || cls->isLearnt()){ //make sure that rather learnt clses are wasted!
						LOG2("Strengthen resolvent after cls reason");
						prev_res_sz = strengthenClsSetConflicting(prev,fp_lit);
						if(stren_cls){cls->setWaste();	LOG2("Delete second reason");}
					}else{
						LOG2("Strengthen both after cls reason");
						prev_res_sz = strengthenClsSetConflicting(cls,p_lit);
						prev->setWaste();
					}
				}else if(resolv_sz == cls_rel){
					LOG2("Strengthen reason");
					prev_res_sz = strengthenClsSetConflicting(cls,p_lit);
				}else{prev_res_sz= 0;}
			}
			asrt(prev_res_sz<=1 || lemma.size()+nbr_pivots==prev_res_sz);
		}
	}
	asrt(false);return false;
}

BCPETEMP
uint32_t BCPECLASS::backtrackLevel(Literal& max){
	asrt(lemma.size() > 0);
	asrt(sstate.vinf.litAssLev(lemma.back()) == decisionLevel());

	max				= FALSE_LIT;
	uint32_t mlev	= 0;
	uint32_t cand;
	for(uint32_t b = lemma.size()-1;b--;){
		if((cand=sstate.vinf.litAssLev(lemma[b]),cand > mlev)){
			mlev = cand; max = lemma[b];} asrt(sstate.vinf.litFalse(lemma[b]));
	}

	asrt(mlev < sstate.vinf.litAssLev(lemma.back()));
	return mlev;
}


//try to minimise the conflict clause
BCPETEMP
void	BCPECLASS::lemmaReduction(const bool only_clear){

	asrt(some_lits.empty()); 	asrt(sstate.vinf.litAssLev(lemma.back()) == decisionLevel());
	if(only_clear || lemma.size() < 2){
		for(uint32_t i=lemma.size();i--;){
			zeroFlags(lemma[i]);
			delLevel(sstate.vinf.litAssLev(lemma[i]));
		}
		return;
	}


	uint32_t i,cp,end; asrt(lemma.size()>1); asrt(asrtAllLitsFlagged(lemma));
	Literal root 			= sstate.vinf.litRoot(lemma[0]);
	end 					= lemma.size()-1;
	const Literal asserting = lemma[end];

	for(cp = i = 0; i != end; ++i){asrt(sstate.vinf.litFalse(lemma[i]));
		Literal lit = lemma[i];
		if(needLitGraphCheck(lit)) { //take literal
			lemma[cp++] = lit; asrt(isSet(lit));
			if(sstate.vinf.litRoot(lit)!=root){root = FALSE_LIT;}
		}else{
			some_lits.push_back(lit);}
	}
	sstate.stats->nbr_lemma_reduct += (end - cp);
	lemma[cp++] = asserting; asrt(cp <= lemma.size()); asrt(isSet(lemma[end]));
	asrt(root != TRUE_LIT);

	//bool change = !some_lits.empty();
	for(i = some_lits.size();i--;) {clearFlag(some_lits[i]);}
	some_lits.clear();

/*
	end = cp;
	for(cp = i = conflict_lemma.begin()+1; i != end; ++i){
		if(needLitShortsCheck(*i)){*cp++ = *i;} else {clearFlag(*i);delLevel(litAssLev(*i));} }
	nbr_lemma_reduct_shrts += (end - cp);
*/

	lemma.resize(cp); 	asrt(asrtAllLitsFlagged(lemma)); asrt(sstate.vinf.litAssLev(lemma.back()) == decisionLevel());

	while(cp--){
		Literal lit = lemma[cp];
		zeroFlags(lit); delLevel(sstate.vinf.litAssLev(lit));	}


	asrt(root != lemma.back()); asrt(root != lemma.back().flipped());
	if(root != FALSE_LIT && lemma.size()>2 ){asrt(sstate.vinf.litFalse(root));
		sstate.stats			->nbr_hyper_bin_lemma_red++;
		lemma		.resize(2);
		lemma[0]	= root;
		lemma[1]	= asserting;
		//change		= true;
	}

	asrt(lit_flags.allCleared());		asrt(some_lits.empty());
	asrt(level_flags.allCleared()); 	asrt(sstate.vinf.litAssLev(lemma.back()) == decisionLevel());
	//return change;
}

/**Checks if the given literal is needed in the conflict clause.
 * If it can be resolved from the current reasons it can be omitted*/
BCPETEMP
bool	BCPECLASS::needLitGraphCheck(const Literal& l){

	if(sstate.vinf.varIsDecision(l.toVar())) return true;

	Literal lit = l;
	uint32_t unmark_top(some_lits.size());
	asrt(tmp_literal_store.empty());
	bool is_nec = false;

	while(true){

		const BDReason& reason = sstate.vinf.varAssignReason(lit.toVar());
		asrt(!reason.isDecision());

		if(reason.isBinReason()){
			is_nec = 	needLitGraphCheckHelper(reason.getBinReason());
		}else{
			asrt(reason.isClsReason() && !reason.isDecision());

			SharedClause* cls 	= sstate.cinf.get(reason.getClsReason()).getCls(); asrt(cls != NULL);
			const Literal lf	= lit.flipped();

			FOR_LITS_STRT(cls,n_lit);

				if(n_lit == lf) continue;
				if((is_nec = needLitGraphCheckHelper(n_lit),is_nec)) break;

			FOR_LITS_END;

		}

		if(is_nec){
			for( ;unmark_top < some_lits.size();++unmark_top)	{clearFlag(some_lits[unmark_top]);}
			for(int i = tmp_literal_store.size();i--;)			{clearFlag(tmp_literal_store[i]);}
			tmp_literal_store.clear();
			return true;
		}


		if(tmp_literal_store.empty()){
			//statistics.total_lemma_shortening += 1;
			return false;
		}

		lit = tmp_literal_store.back(); tmp_literal_store.pop_back();
		some_lits.push_back(lit);
	}
}

/*
BCPETEMP
bool BCPECLASS::learnUnitClses(vector<Literal>& units){

	uint32_t impl = 0;
	for(uint32_t i=0;i!=units.size();++i){
		Literal imp = units[i];
		if(sstate.vinf.litFixTrue(imp)) continue;

		backjumpKeepLevel(0);		asrt(!openBCPImpls());

		if(sstate.vinf.litFalse(imp)) return setInvalid("Analysis implies unit");
		asrt(sstate.vinf.litUkwn(imp));

		if(!enqueue(imp,0) || (bcpLong()!=BCP_SUCCESS)){
			return setInvalid("Analysis implies unit -> bad enqueue");}
		++impl;
	}

	sstate.stats->bin_impl += impl;
	units.clear();
	asrt(!openBCPImpls());
	return true;
}*/

BCPETEMP
void BCPECLASS::wasteAllWatched( const Literal& lit){
	Iter<WtchEle> it = sstate.wtch.getWtchOf(lit);
	WtchEle we;
	while(it.getNext(we)){sstate.cinf.get(we.clp).getCls()->setWaste();}
}

BCPETEMP
void BCPECLASS::clearBins(const Literal& lit){
	Literal otherw;
	WIter<Literal> bi	= sstate.wtch.getWBinsOf(lit);
	if(!sstate.vinf.litFix(lit) && !sstate.vinf.varElim(lit.toVar())){
		while(bi.getNext(otherw)){
			if(!sstate.vinf.litFix(otherw) && !sstate.vinf.varElim(otherw.toVar())){
				bi.keep(otherw);}
		}
	}
	sstate.wtch.finish(lit,bi);
}


BCPETEMP
void BCPECLASS::clearWatches( const Literal& lit, const uint32_t max_lbd){
	LOG3("Clear Watchers of literal "<<lit.toDIMACS()<<" at level "<<decisionLevel());
	asrt(asrtAllUnitsSent());
	clearBins(lit);

	const bool lit_rem		= sstate.vinf.varElim(lit.toVar());
	const bool lit_fix		= sstate.vinf.litFix(lit);
	//const bool del_unusd	= (max_lbd != max32);
	WtchEle we;
	WIter<WtchEle>	wi		= sstate.wtch.getWtchOfW(lit);

	while(wi.getNext(we)){

		ClsWrp& cw			= sstate.cinf.get(we.clp);
		we.other			= cw.getOtherW(lit);

		if(we.other==lit){	//already handled and marked for deletion
			sstate.cinf	.freeCls(we.clp, sstate.s_info->user_bit);	continue;}

		if(we.other < lit){ //already handled without effect
			wi.keep(we);  continue;}

		asrt(lit < we.other); //we have to check for deletion
		const bool locked 	= cw.isLockd();
		cw.setLockd(false);
		if(locked){			wi.keep(we); continue;}

		SharedClause* cls	= cw.getCls();
		if(		lit_rem
			|| (lit_fix && sstate.vinf.litTrue(lit))
			|| sstate.vinf.litFixTrue(we.other)
			|| sstate.vinf.varElim(we.other.toVar())){
			cls->setWaste();}

		if(cls->isWaste() || (cw.getLBD()>max_lbd) ){ asrt(cls->isWaste() || cls->isLearnt());
			cw.invalidateWatchers(); LOG2("Get rid of cls ");} //clause will be deleted by second watcher
		else{
			wi.keep(we);}
	}
	sstate.wtch.finish(lit,wi);
}



BCPETEMP
void BCPECLASS::lockReasons(){
	if(decisionLevel() == 0) return;

	for(uint32_t d = decision_level_end[0]; d < assign_stack.size();++d){
		const BDReason& r 	= sstate.vinf.varAssignReason(assign_stack[d].toVar());
		if(r.isClsReason())	{sstate.cinf.get(r.getClsReason()).setLockd(true);}
	}
}

BCPETEMP
void BCPECLASS::clearWatches(const uint32_t max_lbd){

	sendUnits();

	LOG1("Start clearing watchers"); sanity(sstate.wtch.asrtBinsBothDirs(sstate.vinf));
	for(uint32_t w=0;w!=sstate.stats->nVariables(); ++w){
		const Variable v(w);
		if(!sstate.vinf.varRepl(v)){
			clearWatches(v.toNegLit(),max_lbd);
			clearWatches(v.toPosLit(),max_lbd);
		}
	}

	sstate.cls_reg.cleanWasteAndDeleted();
	beyUP_cands	.clear();
	beyUP_head 	= 0;
	LOG1("Finish clearing watchers"); sanity(sstate.wtch.asrtBinsBothDirs(sstate.vinf));
}



BCPETEMP
void BCPECLASS::clsGC(const uint32_t new_sz){

	if(new_sz >= sstate.stats->nLearntLongClses()){
		//CONSOLE("Shall do gc but too small " << new_sz << " requested but we have " << sstate.stats->nLearntLongClses() << endl);
		return ;
	}
	LOG2("Starting learntsGC with newsz "<< new_sz<<" and current learnts "<<sstate.stats->nLearntLongClses());
	TIME_ON(sstate.stats->gc_time);

	++sstate.stats->nbr_gc;
	lockReasons();
	uint32_t lbd 	= sstate.cinf.getLBDWithRank(new_sz);

	if(lbd<2){lbd = 2;LOG2("Median LBD smaller than 2"); }
	if(lbd != SharedClause::LBD_MAX){sstate.s_info->setLastLBD(lbd);}

	clearWatches(lbd);
	LOG1("Finished learntsGC with "<<sstate.stats->nbr_ll_cls<<" learnts ");
	TIME_OFF(sstate.stats->gc_time);
}






/*
bool BCPECLASS::needLitShortsCheck(const Literal& lit){
	asrt(litFalse(lit));
	WatchedList& wlist = sstate.wtch.get(lit.flipped());


	WatchedList::siter_t siter		= wlist.shortsBegin();
	const WatchedList::siter_t send	= wlist.shortsEnd();

	while(siter != send){asrt(siter<send);

		if(!isSet(wlist.getFirstOfShrt(siter)))	{wlist.incAfterUkwn(siter);continue;}

		if(wlist.shortIsBinary(siter))			{return false;}

		if(isSet(wlist.getSecondOfShrt(siter))) {return false;}
		wlist.incAfterTern(siter);
	}
	return true;
}*/



////////////////////
//Cleaning Clauses








BCPETEMP
template<class CollVisi,class MarkVisi>
void BCPECLASS::collectCommon(Literal lits[], uint32_t lsz,vector<Literal>& common_out){
	sanity(lit_flags.allCleared());
	if(lsz<2){	if(lsz==1){common_out.push_back(lits[0]);} return;}
	LOG3("Start search for common literals of " << lsz);

	CollVisi coll(sstate.vinf); MarkVisi mark(sstate.vinf);
	BFS(lits[--lsz],coll,true,true);
	if(coll.common.empty()){LOG3("No common lits ");return;}

	while(lsz--){	asrt(!coll.common.empty());
		BFS(lits[lsz],mark,true,false);
		uint32_t cp = 0;
		for(uint32_t i=0;i!=coll.common.size();++i){
			if(sstate.vinf.isFlagged(coll.common[i].first)){
				coll.common[cp++]=coll.common[i];LOG3("Keep lit "<<coll.common[i].first.toDIMACS());}}
		clearBFSFlags();
		if(cp == 0){LOG3("No common lits ");return;}
		coll.common.resize(cp);
	}
	asrt(!coll.common.empty());
	LOG3("By now common lits: " << coll.common.size());

	//take only topmost common literals -> others are redundant
	for(uint32_t i=coll.common.size();i--;){setFlag(coll.common[i].first);LOG3("Flag lit"<<coll.common[i].first.toDIMACS());}
	for(uint32_t i=coll.common.size();i--;){
		if(!isSet(coll.common[i].second)){common_out.push_back(coll.common[i].first);}
		else{LOG3("Dont take common " << coll.common[i].first.toDIMACS() << " because predecessor is taken: " << coll.common[i].second.toDIMACS());}
		clearFlag(coll.common[i].first);}

	LOG2("Found common literals: " << common_out.size());
	asrt(!common_out.empty());	sanity(lit_flags.allCleared());
}



//for each target the Visitor functor is called exactly once
//the second argument is the predecessor by which target was reached first -> hence not all edges are called with visitor
//if visit_start is true the start literal is visited with predecessor "INVALID_LIT"

BCPETEMP
template<class Visitor>
void BCPECLASS::BFS(const Literal& start, Visitor& visi, const bool visit_start, const bool clear_flags){
	asrt(!clear_flags || sstate.vinf.asrtAllFlagsClear()); asrt(!sstate.vinf.litRepl(start));
	asrt(!clear_flags || tmp_literal_store.empty()); asrt(!sstate.vinf.litFix(start));
	LOG3("BFS run with start " << start.toDIMACS());
	uint32_t qhead = tmp_literal_store.size();
	if(!visit_start || visi(start,FALSE_LIT).isTrue()){
		sstate.vinf.setFlag(start); tmp_literal_store.push_back(start);}

	Literal target;
	while(qhead != tmp_literal_store.size()){
		Literal src 		= tmp_literal_store[qhead++]; //asrt(!dag.isLitRepl(src));
		Iter<Literal> bi	= sstate.wtch.getBinsOf(src.flipped());

		while(bi.getNext(target)){
			if(sstate.vinf.litFix(target)) {LOG3("target fix "<<target.toDIMACS());continue;}

			asrt(!sstate.vinf.litRepl(target));
			if(sstate.vinf.isFlagged(target)) continue;
			TrBool v_res				= visi(target,src);
			if(v_res.isUnknown()){		qhead=tmp_literal_store.size();break;}
			if(v_res.isTrue()){
				sstate.vinf.setFlag(target);	tmp_literal_store.push_back(target);}
		}
	}

	LOG3("Finishing BFS run with start " << start.toDIMACS());
	if(clear_flags){clearBFSFlags();}
}

BCPETEMP
void BCPECLASS::clearBFSFlags(){
	for(uint32_t i=tmp_literal_store.size();i--;){
		sstate.vinf.clearFlag(tmp_literal_store[i]);}
	tmp_literal_store.clear();
}



BCPETEMP
bool BCPECLASS::oneKnwnTrue(ElimCls* eli){
	asrt(!sstate.vinf.litTrue(eli->ukwn));
	for(uint32_t i=eli->size;i--;){
		Literal l = sstate.vinf.findLitRepl(eli->knwn[i]);

		asrt(!sstate.vinf.varElim(l.toVar())); asrt(!sstate.vinf.litUkwn(l));
		if(sstate.vinf.litTrue(l)) return true;
	}
	return false;
}


BCPETEMP
void BCPECLASS::uneliminate(){

	//reimport eliminated clauses
	uint32_t dec_l = decisionLevel();
	Variable last = TRUE_LIT.toVar();
	for(ElimCls* eli = sstate.s_info->elim.getLast();eli != NULL; eli = eli->getPrev()){
		Literal cand = eli->ukwn; Variable candv(cand.toVar());

		//handle variables with dont care values
		if(candv != last && !sstate.vinf.varKnwn(last)){
			LOG1("Set last because unknown " << last.toDIMACS());
			sstate.vinf.setValueDecision(last.toPosLit(),dec_l);}
		last = candv;


		//ignore value of eliminated variable
		if(sstate.vinf.varElim(candv)){
			sstate.vinf.unsetElim(candv);
			if(!sstate.vinf.varFix(candv)){sstate.vinf.varClearValue(candv);}
		}

		if(eli->size==0) continue; //dummy just to uneliminate all eliminated


		if(sstate.vinf.varKnwn(candv)){asrt(sstate.vinf.litTrue(cand)||oneKnwnTrue(eli));	continue;}

		if(!oneKnwnTrue(eli)){sstate.vinf.setValueDecision(cand,dec_l);}
	}

}


BCPETEMP
Literal BCPECLASS::getModelVal(const Variable& v){
	Literal pr 	= sstate.vinf.findLitRepl(v.toPosLit()); asrt(!sstate.vinf.litRepl(pr));
	return  sstate.vinf.varPrevValWasTrue(pr.toVar()) == pr.isPos() ?
			v.toPosLit() : v.toNegLit();

}


BCPETEMP
void BCPECLASS::printModel(){

	uneliminate();
	printf("\nv ");
	for(uint i=1; i != sstate.stats->nVariables();++i){
		Literal val	= getModelVal(Variable(i));
		printf("%d ", val.toDIMACS());
	}
	printf("0 \n");
}


BCPETEMP
void BCPECLASS::getModel(vector<bool>& var_vals){
	//asrt(assign_stack.size()==sstate.stats->nVariables());
	uneliminate();
	var_vals.clear();
	var_vals.reserve(sstate.stats->nVariables());
	for(uint i=0; i != sstate.stats->nVariables();++i){
		Literal val = getModelVal(Variable(i));
		var_vals.push_back(val.isPos());}
}


BCPETEMP
void BCPECLASS::printDIMACS(const string& file,const TrBool& result){
	asrt(!file.empty());
	std::ofstream out(file.c_str());

	if(!result.isTrue()){backjumpKeepLevel(0);}

	for(uint i=1; i != assign_stack.size();++i){
		out << assign_stack[i].toDIMACS() << " 0" << endl;}

	for(uint32_t i=2; i!= (sstate.stats->nVariables()*2);++i){
		const Literal lit	(i);
		if(sstate.vinf.litFix(lit)) continue;
		Literal other		= sstate.vinf.findLitRepl(lit);

		if(other!=lit) 		{
			if(lit < other){
				out << lit.flipped().toDIMACS() << " " << other.toDIMACS() << " 0" << endl;}
			out << "c repl: " << lit.toDIMACS() << " by " << other.toDIMACS() << endl;
			continue;
		}

		Iter<Literal> iter = sstate.wtch.getBinsOf(lit);
		while(iter.getNext(other)){  asrt(other != FALSE_LIT);
			if(lit < other){
				out << lit.toDIMACS() << " " << other.toDIMACS() << " 0" << endl;}
		}
	}


	for(ClsPointer iter = sstate.cinf.lastValid();!iter.isNull();iter=sstate.cinf.nextValidBefore(iter)){
		SharedClause* cls	= sstate.cinf.get(iter).getCls();
		if(cls->isLearnt())	continue;

		FOR_LITS_STRT(cls,lit);
			out << lit.toDIMACS() << " ";
		FOR_LITS_END;
		out		<< "0" << endl;
	}
}





BCPETEMP
bool BCPECLASS::asrtCheckSatisfied(SharedClause* cls) const{
	FOR_LITS_STRT(cls,lit);
	    if(sstate.vinf.litTrue(lit)) return true;
	FOR_LITS_END;
    return false;
}


BCPETEMP
bool	BCPECLASS::asrtAllLitsFlagged(const vector<Literal>& lits){
	for(uint i = lits.size();i--;){
		if(!isSet(lits[i])) return false;
	}
	return true;
}

BCPETEMP
bool	BCPECLASS::asrtClsUnit(SharedClause* cls,const Literal& l) const{
	bool ok = true;
	if(!sstate.vinf.litUkwn(l)){
		ASRT_PRNT("Literal not free: " << l.toDIMACS() << endl);ok=false;}

	FOR_LITS_STRT(cls,lit);

		if(lit != l && !sstate.vinf.litFalse(lit)){
			ASRT_PRNT("Literal not false: " << lit.toDIMACS() << " but " << (sstate.vinf.litUkwn(lit)?" unknown":" true")<<endl);ok=false;}

	FOR_LITS_END;

	return ok;
}

/*
BCPETEMP
bool	BCPECLASS::asrtClsSorted(SharedClause* cls) const{
	Literal last = TRUE_LIT;
	FOR_LITS_STRT(cls,lit);
		if(last>lit) return false;
		last = lit;
	FOR_LITS_END;

	return true;
}


BCPETEMP
bool	BCPECLASS::asrtSortedAndFlagged(const Literal lits[],const uint32_t lsz)const{
	if(lsz==0){ASRT_PRNT("Check with empty clause" << endl); return false;}
	Literal last = lits[0];
	for(uint i=0;i!=lsz;++i){
		if(last.idx()>lits[i].idx()) {ASRT_PRNT("NOT SORTED " << endl);	return false;}
		last = lits[i];
		if(!isSet(last)){ASRT_PRNT("NOT all Flagged " << endl);	return false;}
	}
	return true;
}*/

BCPETEMP
bool	BCPECLASS::asrtBinPropagatedIf(const Literal& lit){
	if(!sstate.vinf.litTrue(lit)) return true;
	Literal repl 	= sstate.vinf.findLitRepl(lit);
	uint32_t lev	= sstate.vinf.litAssLev(lit);
	if(repl != lit){
		asrt(repl != FALSE_LIT);
		if(!sstate.vinf.litTrue(repl) || sstate.vinf.litAssLev(repl)>lev){
			ASRT_PRNT("Lit " << lit.toDIMACS() <<" (lev:"<<lev<< ") repl by " <<
					repl.toDIMACS() <<" (lev:"<<sstate.vinf.litAssLev(repl)<< ") not well propagated" << endl);
			return false;
		}
	}else{
		Iter<Literal> bi	= sstate.wtch.getBinsOf(lit.flipped());
		while(bi.getNext(repl)){
			asrt(repl != FALSE_LIT);
			if(!sstate.vinf.litTrue(repl)||sstate.vinf.litAssLev(repl)>lev){
				string ov = sstate.vinf.litUkwn(repl) ? "(unknown)" : (sstate.vinf.litTrue(repl)?"(true)":"(false)");
				ASRT_PRNT("Lit " << lit.toDIMACS() <<" (lev:"<<lev<< ") implies " <<
						repl.toDIMACS() << ov <<" (lev:"<<sstate.vinf.litAssLev(repl)<< ") not well propagated" << endl);
				return false;
			}
		}
	}
	return true;
}


BCPETEMP
bool BCPECLASS::asrtResolventSize(uint32_t sz) const{
	const uint32_t lev		= decisionLevel();
	uint32_t count			= lemma.size();

	for(uint32_t i=assign_stack.size();i--;){
		Literal flit = assign_stack[i].flipped();
		if(isSet(flit)){
			if(sstate.vinf.varAssignLevel(flit.toVar())<lev) break;
			++count;}
	}
	if(sz == count){ return true;}

	cout << "c Size of Resolvent was assumed to be " << sz << " but is " << count << endl;
	return false;
}


BCPETEMP
void BCPECLASS::asrtPrintCurrResolvent() const {
	vector<Literal> curr 	= lemma;
	const uint32_t lev		= decisionLevel();

	for(uint32_t i=assign_stack.size();i--;){
		Literal flit = assign_stack[i].flipped();
		if(isSet(flit)){
			if(sstate.vinf.varAssignLevel(flit.toVar())<lev) break;
			curr.push_back(flit);}
	}

	std::sort(curr.begin(),curr.end());
	std::cout << "c Current lemma at " << lev << ": ";
	for(uint32_t i=0;i!=curr.size();++i){
		std::cout << curr[i].toDIMACS() << " ";//<< " [" << sstate.vinf.varAssignLevel(curr[i].toVar()) << "] ";
	}
	std::cout << endl;
}


BCPETEMP
bool BCPECLASS::asrtNoClsContainsReplOrElim(){
	bool ok = true;
	for(ClsPointer iter = sstate.cinf.lastValid();!iter.isNull(); iter = sstate.cinf.nextValidBefore(iter)){
		SharedClause* cls	= sstate.cinf.get(iter).getCls();

		FOR_LITS_STRT(cls,lit);
			if(sstate.vinf.varElim(lit.toVar())){ ok = false;
				ASRT_PRNT("cls contains eliminated lit " << lit.toDIMACS());
			}
			if(sstate.vinf.litRepl(lit)){ ok = false;
				ASRT_PRNT("cls contains replaced lit " << lit.toDIMACS());
			}

		FOR_LITS_END;
	}

	return ok;
}


BCPETEMP
bool BCPECLASS::asrtAllLinkedClsesSAT(const vector<bool>& var_vals){
	bool ok=true; Literal oth; WtchEle we;

	for(uint32_t i=var_vals.size()*2;i--;){
		Literal l(i); Variable v = l.toVar();
		bool pol 		= var_vals[v.idx()];
		bool sat_all	= (pol == l.isPos());

		if(!sat_all){
			Iter<Literal> bi = sstate.wtch.getBinsOf(l);
			while(bi.getNext(oth)){
				Variable ov = oth.toVar();
				if(ov.idx() >= var_vals.size()){ ok=false;
					ASRT_PRNT("var too big " << ov.toDIMACS());
				}else if(var_vals[ov.idx()] != oth.isPos()){ ok=false;
					ASRT_PRNT("binary clause not sat" << l.toDIMACS() << " " << oth.toDIMACS());
				}
			}
		}


		Iter<WtchEle> wi = sstate.wtch.getWtchOf(l);
		while(wi.getNext(we)){
			SharedClause* cls = sstate.cinf.getCls(we.clp);

			bool sat_this = false;
			FOR_LITS_STRT(cls,lit);
			Variable ov = lit.toVar();
			if(ov.idx() >= var_vals.size()){ ok=false;
				ASRT_PRNT("var too big " << ov.toDIMACS());
			}else if(var_vals[ov.idx()] == lit.isPos()){
				sat_this = true; break;
			}
			FOR_LITS_END;

			if(sat_this){ std::cerr << "cls verified: " << ((uint64_t)cls) << std::endl; continue;}

			ok=false;
			ASRT_PRNT("does not sat the following clause ");
			FOR_LITS_STRT(cls,lit);
				std::cerr << lit.toDIMACS() << " ";
			FOR_LITS_END;
			std::cerr << " /" << endl;
		}
	}

	return ok;
}




#endif /* BcpGraph_HPP_ */
