/*
 * MPRun.hpp
 *
 *  Created on: Oct 26, 2009
 *      Author: kottlers
 */

#ifndef OMPRUN_HPP_
#define OMPRUN_HPP_
#define omprun_hpp_version "$Revision: 1.4 $"

#include "GlobalSettings.hpp"
#include "SolverFactory.hpp"

using std::cout; using std::endl;

template<class SOLVERS>
inline int ompRun(SOLVERS& solver_handler,GlobalSettings& global_settings, IFStream& in){

	int return_val = EXIT_UNKNOWN;
	if(!solver_handler.parseDIMACS(in)) {
		return_val= solver_handler.printResult(UserID(0),global_settings.cnf_file_name,FALSE_TrBool);
		exit(return_val);
	}
	in.close();

	solver_handler.initSharing();

	//CONSOLE("MEMORY USAGE: " <<  ((double)usedMemoryBytes()/(double)MEGA) << endl);

	TrBool result;
	#pragma omp parallel num_threads(global_settings.nbr_threads) private(result)
	{
		UserID uid(omp_get_thread_num());

	    LOG2("Here we go");

	    result = solver_handler.invokeSolver(uid);

	    LOG2("Terminating");

	    if(solver_handler.canPrintFinal()) //#pragma omp single
		{
			solver_handler.stopPrinting();
			LOG1("Computation took " << cpuUsrTime() << " seconds ");

			solver_handler.stopAllASAP();

			if(global_settings.print_console) 	solver_handler.printStatsConsole(uid,result);
			return_val= solver_handler.printResult(uid,global_settings.cnf_file_name,result);

			if(!global_settings.csv_file_name.empty()){
				solver_handler.printCSV(uid,global_settings.csv_file_name,result);}

			/*if(!global_settings.dimacs_out.empty()){
				solver_handler.printDIMACS(uid,global_settings.dimacs_out,result);}*/

			#ifdef NDEBUG
				exit(return_val);
			#endif
		}

		solver_handler.checkOut(uid);
	}
	return return_val;
}






inline int ompRun(GlobalSettings& global_settings){

	IFStream in(global_settings.cnf_file_name.c_str());
	if(in.isBad()){
		std::cerr << "c Could not open CNF file " <<  global_settings.cnf_file_name << endl;
		global_settings.printUsage();
		return EXIT_FAILURE;
	}

	Logging::setNumberThreads(global_settings.nbr_threads,"./");

	if(global_settings.print_console) Satistics::printParaHead(global_settings.nbr_threads);

	int ret = 0;
	switch(global_settings.nbr_threads){
	case 1:{
		std::cerr << "c It's possible but not recommended to only use one thread with these settings" <<endl;
		SingleSolver s(global_settings);
		ret = ompRun(s,global_settings, in);
		break;}

	default:{
	//case 4:{
		/*std::cerr << "c special setting for crafted" <<endl;
		ExplicitCraftedSolver s(global_settings.nbr_threads);
		s.initTypes(global_settings);
		ret = ompRun(s,global_settings,in);
		break;}*/


		//CONSOLE("Cannot handle " << global_settings.nbr_threads << " Threads " << endl);}
		//std::cerr << "c default explicit solver" <<endl;
		ExplicitSolver s(global_settings.nbr_threads);
		s.initTypes(global_settings);
		ret = ompRun(s,global_settings,in);
	}
	}


	Logging::closeThreads();
	return ret;
}





#endif /* OMPRUN_HPP_ */
