#include "SCC.hpp"
#include <algorithm>

const char*  scc_cpp_version(){return "$Revision: 1.5 $";}

typedef list<Literal>::iterator liter;
typedef list<Literal>::const_iterator cliter;



bool SCC::sccPushStack(const Literal& lit){ asrt(!sstate.vinf.litKnwn(lit)); asrt(!sstate.vinf.litElim(lit));

	Iter<Literal> it		= sstate.wtch.getBinsOf(lit.flipped());
	if(!it.hasNext()){		doFini(lit); return false;}

	const uint32_t clos_sz 	= closure.size();
	const uint32_t li_idx 	= scc_collect.size();
	doStck					(lit);
	sccSetCollIdx			(lit,li_idx);
	scc_stack	.push_back	(StackEle(li_idx,clos_sz));
	scc_collect	.push_back	(SCCCollect(lit,li_idx)); 	asrt(scc_collect[li_idx].lit==lit);

	uint32_t clsr; Literal chld;
	while(it.getNext(chld)){ asrt(!sstate.vinf.litElim(chld));

		//remove known lits
		if(sstate.vinf.litKnwn(chld)){doFini(chld);continue;}

		//check if shortcut exists
		if(!stStck(chld) && (clsr=rdntClosureIdx(chld),clsr!=max32)){
			if(clsr>=clos_sz){LOG2("Remove duplicate bin to " << chld.toDIMACS()) ;continue;} //double bin entry

			closure[clsr].vld = 0;
			LOG2("Mark shortcut to "  << chld.toDIMACS());
			sstate.stats->nbr_scc_shortcuts++;
		}

		rdntSetClosureIdx(chld,closure.size());//parent shows to 'lit'
		closure	.push_back(Closure(chld,true)); asrt(chld.idx() == closure[rdntClosureIdx(chld)].lit);
	}
	LOG2("Closure of " << lit.flipped().toDIMACS() << " has size " << closure.size()-clos_sz);

	return true;
}



void SCC::sccPopCollect(Literal& lit){
	lit = scc_collect.back().lit; asrt(stStck(lit));
	doFini(lit);
	scc_collect.pop_back();
}


void SCC::sccPopStack(const Literal& top_lit,vector<Message>* rem_bin_out){
	asrt(!scc_stack.empty());
	StackEle& top 		= scc_stack.back(); asrt(top.coll_top>=scc_collect.size() || scc_collect[top.coll_top].lit == top_lit);
	const Literal ft	= top_lit.flipped();
	LOG2("Pop " << ft.flipped().toDIMACS() << ". Nbr lits in closure of " << ft.toDIMACS()<<": "
				<< closure.size()-top.child_strt << " orig: " << sstate.wtch.nbrBins(ft));

	asrt(sstate.wtch.nbrBins(ft)>=closure.size()-top.child_strt);
	Range<Literal> r = sstate.wtch.getBinsRangeOf(ft);
	for(uint32_t i=top.child_strt;i<closure.size();++i){
		Closure& clsr = closure[i]; Literal cld(clsr.lit); asrt(cld != ft);
		if(clsr.vld || stStck(cld.flipped())){ //ensure that no binaries are removed where closure is active
			sstate.wtch.replaceNextBin(r,cld);} //keep it
		else{
			LOG2("Omit shortcut " << ft.toDIMACS() << " and " << cld.toDIMACS());
			sstate.wtch.removeBinPart(cld,ft);
			if(rem_bin_out != NULL){
				rem_bin_out->push_back(Message::newBinRemove(cld,ft));}
		}
		rdntSetClosureIdx(cld,max32);
	}
	sstate.wtch.finishBinReplace(ft,r);

	closure		.resize(top.child_strt);
	scc_stack	.pop_back();
}



void SCC::sccStart(const Literal& lit,list<list<Literal> >& sccs, vector<Literal>& units_out,vector<Message>* remv_bins_out){
	asrt(sstate.vinf.litUkwn(lit)); asrt(scc_stack.empty());  asrt(scc_collect.empty());
	Literal s,min,top_lit;
	sccPushStack(lit);

	while(!scc_stack.empty()){
		const uint32_t top_idx 	= scc_stack.size()-1;	asrt(scc_stack.size()<=scc_collect.size());
		StackEle& top_s			= scc_stack[top_idx];   asrt(top_s.coll_top < scc_collect.size());
		SCCCollect& top_c		= scc_collect[top_s.coll_top]; asrt(top_c.root<scc_collect.size());

		//element ready - all children finished
		if(top_s.next_child >= closure.size()){ 	asrt(top_s.next_child==closure.size());LOG2("top element ready");
			top_lit 			= top_c.lit;

			if(top_c.root < top_s.coll_top){				asrt(top_idx>0);LOG2("top lit belongs to scc with root above");
				SCCCollect& prnt	= getCollIfoByStackIdx(top_idx-1);
				if(top_c.root<prnt.root){
					prnt.root=top_c.root;					LOG2("Set root of parent " << prnt.lit.toDIMACS() << " to " << prnt.root);}

			//this is the root of an scc -> take its elements if more than one
			}else{	asrt(top_c.root==top_s.coll_top);
				const uint32_t sz 	= scc_collect.size() - top_c.root;	LOG2("Found root of scc with size " << sz); asrt(sz>0);
				sccPopCollect(s);

				if(sz != 1){
					sccs.push_back(list<Literal>());
					list<Literal>& scc = sccs.back();
					scc.push_front(min=s); 							LOG2("found scc: " << s.toDIMACS());
					while(scc_collect.size()>top_c.root){
						sccPopCollect(s);							LOG2(s.toDIMACS());
						if(s<min){	scc.push_front(min=s);}
						else{		scc.push_back(s);}
					}
				}
			}
			sccPopStack(top_lit,remv_bins_out);

		//element not ready yet
		}else{asrt(top_s.next_child < closure.size());

			do{
				Literal chld (closure[top_s.next_child++].lit);


				if(stStck(chld.flipped())){
					sstate.stats->nbr_scc_units++; LOG2("Unit impl in scc " << chld.toDIMACS());
					units_out.push_back(chld); //chld.flipped implies chld
				}

				if(stStck(chld)){
					LOG2("child " <<chld.toDIMACS()<<" is stacked perhaps adapt root"); asrt(sccCollIdx(chld)<scc_collect.size());

					uint32_t c_root = scc_collect[sccCollIdx(chld)].root;
					if(c_root<top_c.root){top_c.root=c_root;}

				}else if(!stFini(chld)){//move forward
					LOG2("child " <<chld.toDIMACS()<<" put on stack");
					if(sccPushStack(chld)){break;}
				}
			}while(top_s.next_child < closure.size());
		}
	}
}



void SCC::computeSCCs(list<list<Literal> >& sccs_out,vector<Literal>& units_out,vector<Message>* remv_bins_out){
	uint32_t tmp = ST_FINI; ST_FINI = ST_CLEAR; ST_CLEAR = tmp;
	asrt(asrtAllState(ST_CLEAR));

	vector<Literal> do_later;
	int32_t t,mx,inc;
	if(ST_FINI==0){ //just do it in different orders
		t = 0; mx = sstate.cls_reg.lit_space.size(); inc = 1;}
	else{
		mx = -1 ; t = sstate.cls_reg.lit_space.size()-1; inc = -1;}

	for(;t!=mx;t += inc){
		Literal lit(t);
		if(sstate.vinf.litKnwn(lit) || sstate.vinf.litElim(lit)){
			doFini(lit); continue;}

		if(!stFini(lit)){
			if(checkRoot(lit)){sccStart(lit,sccs_out,units_out,remv_bins_out);}
			else{do_later.push_back(lit);}
		}
	}

	for(uint32_t t=0;t!=do_later.size();++t){
		Literal lit = do_later[t]; asrt(!sstate.vinf.litKnwn(lit)); asrt(!sstate.vinf.litElim(lit));

		if(!stFini(lit)){
			sccStart(lit,sccs_out,units_out,remv_bins_out);}
	}

	sanity(asrtAllState(ST_FINI));
}






bool SCC::asrtAllState(const uint32_t& v) const{
	bool ok = true;
	for(uint32_t i=0;i!=sstate.cls_reg.lit_space.size();++i){
		Literal lit(i);
		uint64_t val = sstate.cls_reg.lit_space[i] & MSK_ST;
		if(v != val){ok=false;
			ASRT_PRNT("Literal " << lit.toDIMACS() << " has bad state " << val << " should have " <<v);
		}
	}

	return ok;
}

