/*
 * WatchAndReason.hpp
 *
 *  Created on: Aug 24, 2009
 *      Author: kottlers
 */

#ifndef WATCHANDREASON_HPP_
#define WATCHANDREASON_HPP_
#define variableinfos_hpp_version      "$Revision: 1.8 $"

#include "GlobalSettings.hpp"
#include "ClauseInfos.hpp"


struct BDReason{
	static const uint32_t MAX_LBD = 15;
protected:
	static const uint32_t DEC 	= 0;
	static const uint32_t CLS 	= 1;
	static const uint32_t BIN 	= 2;
	static const uint32_t ELIM	= 3;
	static const uint32_t REPL	= 4;
	static const uint32_t ELIS	= 5;


	//state the kind of reason
	uint64_t	type 	: 3; //Decision / Cls / Binary / Elim / Replaced

	uint64_t lbd	 	: 4;

	//for binary reasons we keep the literal
	//else a pointer to the clause that became unit under an assignment
	uint64_t reason 	: 31;

	//decision level
	uint64_t level		: 26;



public:
	BDReason():type(DEC),lbd(MAX_LBD),reason(max32>>1),level(max32>>6){}
	bool		isBinReason()		const	{ return type == BIN;}
	bool		isClsReason()		const	{ return type == CLS;}
	bool		isDecision()		const	{ return type == DEC;}
	bool		isElim()			const	{ return type == ELIM;}
	bool		isRepl()			const	{ return type == REPL;}
	bool		isElimSoon()		const	{ return type == ELIS;}
	bool		isReason(const ClsPointer& clp)const
				{return isClsReason() && clp.idx() == reason;}

	Literal		getBinReason()		const{ asrt(isBinReason());return Literal(reason);}
	ClsPointer	getClsReason()  	const	{ asrt(isClsReason());return ClsPointer(reason);}
	uint32_t	getLevel()			const	{return level;}
	//Literal	getDominator()		const	{return domElim;}

	void 		setBinReason(const uint32_t lev, const Literal& l){asrt(level!=0); asrt(!isElim());asrt(!isRepl());
				type = BIN; lbd=MAX_LBD; reason= l.idx(); level=lev; }
	void 		setClsReason(const uint32_t lev, const ClsPointer& clp, uint32_t LBD){asrt(level!=0); asrt(!isElim());asrt(!isRepl()); asrt(LBD<=MAX_LBD);
				type = CLS; lbd=LBD; reason = clp.idx(); level=lev;}
	void 		setDecision(const uint32_t lev){asrt(level!=0); asrt(!isElim());asrt(!isRepl());
				type = DEC; lbd=MAX_LBD; level=lev;}
	void			setElim(){asrt(!isRepl()); type= ELIM;}
	void			setElimSoon(){type= ELIS;}
	void			unsetElim(){asrt(!isRepl()); type= DEC;}
	void			setRepl(const Literal& by){asrt(!isElim());reason = by.idx(); type = REPL;}
	Literal		getRepl() const {asrt(isRepl()); return Literal(reason);}
	uint32_t		getLBD() const {return lbd;}
	//void		setDominator(const Literal& l){dom=l;}

	//void		setLevel(uint32_t m){asrt(level != 0); level=m;}
};



/*

//since we differ between binary and ternary clauses we also have to differ for the assignment reasons
//They are needed to analyse the reason for conflicting assignments
class BTReason:public BReason{

	static const uint32_t TERN 	= 3;

	//for ternary reasons we keep a literal
	Literal 	lit;

public:
	BTReason():BReason(),lit(INVALID_LIT){}
	bool		isTernReason()		const	{ return type == TERN;}

	Literal		getShortSecondReason()const	{ asrt(isTernReason());return lit;}

	void 		setTernReason(const uint32_t lev, const Literal& l1, const Literal& l2){asrt(level!=0);
				level=lev; type = TERN; reason=l1.idx(); lit = l2;}

};

*/


class VarVal{
	TwoBitsVec<Variable> vals;
public:
	VarVal():vals(){}
	VarVal(const VarVal& vls,const bool deep):vals(vls.vals,deep){}
	uint32_t 		count() const{return vals.getSize();}
	/*bool			takeThisIf(const VarVal& v){
					if(v.vals.getSize() == vals.getSize()){ return false;}
					vals = v.vals; return true;}*/
	void			ensureVarVals(uint32_t sz){
					asrt(sz > vals.getSize());
					vals.ensureSize(sz,true);}

	/*void			initWith(const VarVal& from){
					vals = from.vals;}*/


	bool			varHasValue(const Variable& v, bool b)const{
					return vals.hasVal(v,2+(uint32_t)b);}

	bool			varPrevValWasTrue(const Variable& v)const {
					return vals.secSet(v);}

	void			litSetPrevVal(const Literal& l){
					l.isPos() ? vals.setSec(l.toVar()) : vals.clearSec(l.toVar());}

	bool			litPrevValWas(const Literal& l)const {
					return vals.secSet(l.toVar()) == l.isPos();}

	void			varClearValue(const Variable& v){
					vals.clearFirst(v);}

	void			varClearValueFlipPrev(const Variable& v){
					vals.setVal(v,!vals.secSet(v));}

	bool			varKnwn(const Variable& v)const{
					return vals.firstSet(v);}

		//Literals
	bool			litTrue(const Literal& l)const 	{return varHasValue(l.toVar(),l.isPos());}
	bool			litFalse(const Literal& l)const	{return varHasValue(l.toVar(),l.isNeg());}
	bool			litUkwn(const Literal& l)const	{return !varKnwn(l.toVar());}
	bool			litKnwn(const Literal& l)const	{return varKnwn(l.toVar());}

	void			setValue(const Variable& var, const bool& pos){
					vals				.setVal(var,2+((uint32_t)pos));
					asrt(varKnwn(var));	}

	bool			setValueChng(const Variable& var, const bool& pos){
				uint64_t prev = vals.setGetVal(var,2+((uint32_t)pos));asrt(varKnwn(var)); asrt(prev < 2);
				return prev != pos;}

	void			setValue(const Literal& lit){
					setValue(lit.toVar(), lit.isPos());}

};




//template<class Reason>
class BinVarInfo{

	struct Root_XFS{
		uint32_t flag 	: 1;
		uint32_t lit	: 31;
		explicit Root_XFS(uint32_t i):flag(0),lit(i){}
	};

	VarVal&						values;
	einfo<BDReason,Variable>	reasons;
	einfo<Root_XFS,Literal>		roots;

	BinVarInfo(const BinVarInfo& v):values(v.values){asrt(false);}
	BinVarInfo& operator=(const BinVarInfo& v){asrt(false);return *this;}
public:
	BinVarInfo(VarVal& v):values(v){};
	uint32_t varsCount() const {return reasons.size();}

	void	ensureSize(uint32_t sz){
			asrt(sz > values.count());
			reasons			.ensureSize(sz);
			values    		.ensureVarVals(sz);
			roots			.ensureSizeExplicit(sz<<1); asrt(Literal(roots[Literal(sz)].lit)==Literal(sz));	}

	void	createClone(const BinVarInfo& clone){
			asrt(reasons.capa() <100);
			reasons = clone.reasons;
			roots	= clone.roots;
			values	= clone.values;
	}


	bool	varHasValue(const Variable& v, bool b) const {
			return values.varHasValue(v,b);}

	bool	varPrevValWasTrue(const Variable& v) const {
			return values.varPrevValWasTrue(v);}

	bool	litPrevValWas(const Literal& l) const{
			return values.litPrevValWas(l);}

	void	varClearValue(const Variable& v){
			values.varClearValue(v);}

	//void	varClearValueFlipPrev(const Variable& v){
	//		values.setVal(v,!values.secSet(v));}

	bool	varKnwn(const Variable& v) const {
			return values.varKnwn(v);}

	uint32_t varAssignLevel(const Variable& v) const {
			return reasons[v].getLevel();}

	bool 	varFix(const Variable& v) const {
			return reasons[v].getLevel()==0;}

	bool	varIsDecision(const Variable& v) const {
			return reasons[v].isDecision();}

	bool	varHasReason(const Variable& v,const ClsPointer& clp) const{
			return reasons[v].isReason(clp);}

	bool	varIsBinReason(const Variable& v) const{
			return reasons[v].isBinReason();}

	const BDReason& varAssignReason(const Variable& v) const{
			return reasons[v];}

	/*const Literal varDominator(const Variable& v) const{
			return reasons[v].getDominator();}*/

	Literal	varGetBinReason(const Variable& v) const{
			asrt(varIsBinReason(v));
			return reasons[v].getBinReason();}

	/*Literal	varGetShortSecReason(const Variable& v) const{
			asrt(varIsShortReason(v));
			return reasons[v].getShortSecondReason();}*/

	void	setValueDecision(const Literal& lit, uint32_t lev){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			const Literal lf	= lit.flipped();
			values				.setValue(lit);
								asrt(varKnwn(lit.toVar()));asrt(varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setDecision(lev);
			roots[lf].lit		= lf.idx();
	}

	void	setValueReason(const Literal& lit, uint32_t lev,const ClsPointer& clp,const uint32_t& lbd){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			const Literal lf	= lit.flipped();
			values				.setValue(lit);
								asrt(varKnwn(lit.toVar()));asrt(varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setClsReason(lev,clp,lbd);
			roots[lf].lit		= lf.idx();}

	void	setValueReason(const Literal& lit, uint32_t lev,const Literal& r1, const Literal& root){
			const Variable var 	= lit.toVar(); asrt(reasons.inRange(var));
			values				.setValue(lit);
								asrt(varKnwn(lit.toVar()));asrt(varPrevValWasTrue(lit.toVar())==lit.isPos());
			reasons[var]		.setBinReason(lev,r1);
			roots[lit.flipped()].lit = root.idx();}


	//Literals
	bool				litTrue(const Literal& l)		const {return varHasValue(l.toVar(),l.isPos());}
	bool				litFalse(const Literal& l)		const {return varHasValue(l.toVar(),l.isNeg());}
	bool				litUkwn(const Literal& l)		const {return !varKnwn(l.toVar());}
	bool				litKnwn(const Literal& l)		const {return varKnwn(l.toVar());}
	bool				varElim(const Variable& v)const {return reasons[v].isElim();}
	bool				litElim(const Literal& l)const  {return varElim(l.toVar());}
	bool				litElimSoon(const Literal& l)const  {return reasons[l.toVar()].isElimSoon();}
	void				setElimSoon(const Variable& v)	{reasons[v].setElimSoon();}
	void				setElim(const Variable& v)	 {asrt(!reasons[v].isElim());reasons[v].setElim();
																LOG1("Set var elim: "<<v.toDIMACS());}
	void				setRepl(const Literal& who, const Literal& by){
														reasons[who.toVar()].setRepl(who.isPos() ? by : by.flipped());}
	Literal				getRepl(const Literal& who) const{Literal by = reasons[who.toVar()].getRepl();
														return who.isPos()? by : by.flipped();}

	bool				varRepl(const Variable& v)const	{return reasons[v].isRepl();}
	bool				litRepl(const Literal& l)const	{return reasons[l.toVar()].isRepl();}
	inline Literal		findLitRepl(Literal l) const;

	void				unsetElim(const Variable& v)	 {reasons[v].unsetElim();}

	const uint32_t		litAssLev(const Literal& l)		const {return varAssignLevel(l.toVar());}
	//const Literal		litRoot(const Literal& l)		const {asrt(litFalse(l));return roots[l];}
	const Literal		litRoot(const Literal& l)		const {return Literal(roots[l].lit);}
	const bool			litRootUkwn(const Literal& l)	const {return !varKnwn(litRoot(l).toVar());}
	bool 				litFix(const Literal& l) 		const {return varFix(l.toVar());}
	bool 				litFixTrue(const Literal& l) 	const {return varFix(l.toVar()) && litTrue(l);}
	bool 				litFixFalse(const Literal& l) 	const {return varFix(l.toVar()) && litFalse(l);}
	void				setFlag(const Literal& l)			  {roots[l].flag = 1;}
	void				clearFlag(const Literal& l)			  {roots[l].flag = 0;}
	bool				isFlagged(const Literal& l)		const {return roots[l].flag;}
	void				setRootVal(const Literal& l, const uint32_t val){ roots[l].lit = val;}
	uint32_t			getRootVal(const Literal& l) 	const {return roots[l].lit;}

	inline bool 		asrtAllFlagsClear();

};

Literal	BinVarInfo::findLitRepl( Literal l) const{
	while(litRepl(l)){l = getRepl(l);}
	return l;
}



bool BinVarInfo::asrtAllFlagsClear(){
	for(uint32_t i=roots.size();i--;){
		if(roots.byIndex(i).flag){
			ASRT_PRNT("Literal "<< Literal(i).toDIMACS() << " is still flagged");
			return false;}
	}
	return true;
}


//typedef VarInfo<BDReason> BinVarInfo;
//typedef VarInfo<BTReason> BinTernVarInfo;







#endif /* WATCHANDREASON_HPP_ */
